#!/usr/bin/env python
###############################################################################
#                                                                             #
#    cluster.py                                                               #
#                                                                             #
#    A collection of classes / methods used when clustering contigs           #
#                                                                             #
#    Copyright (C) Michael Imelfort                                           #
#                                                                             #
###############################################################################
#                                                                             #
#          .d8888b.                                    888b     d888          #
#         d88P  Y88b                                   8888b   d8888          #
#         888    888                                   88888b.d88888          #
#         888        888d888 .d88b.   .d88b.  88888b.  888Y88888P888          #
#         888  88888 888P"  d88""88b d88""88b 888 "88b 888 Y888P 888          #
#         888    888 888    888  888 888  888 888  888 888  Y8P  888          #
#         Y88b  d88P 888    Y88..88P Y88..88P 888 d88P 888   "   888          #
#          "Y8888P88 888     "Y88P"   "Y88P"  88888P"  888       888          #
#                                             888                             #
#                                             888                             #
#                                             888                             #
#                                                                             #
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Michael Imelfort"
__copyright__ = "Copyright 2012/2013"
__credits__ = ["Michael Imelfort"]
__license__ = "GPL3"
__version__ = "0.2.6"
__maintainer__ = "Michael Imelfort"
__email__ = "mike@mikeimelfort.com"
__status__ = "Alpha"

###############################################################################

from sys import stdout

from colorsys import hsv_to_rgb as htr
import matplotlib.pyplot as plt
from pylab import show
from numpy import (abs as np_abs,
                   allclose as np_allclose,
                   append as np_append,
                   arange as np_arange,
                   argmax as np_argmax,
                   argsort as np_argsort,
                   around as np_around,
                   array as np_array,
                   bincount as np_bincount,
                   concatenate as np_concatenate,
                   copy as np_copy,
                   cos as np_cos,
                   delete as np_delete,
                   finfo as np_finfo,
                   hypot as np_hypot,
                   inf as np_inf,
                   log as np_log,
                   log10 as np_log10,
                   max as np_max,
                   mean as np_mean,
                   median as np_median,
                   min as np_min,
                   ones as np_ones,
                   pi as np_pi,
                   prod as np_prod,
                   reshape as np_reshape,
                   seterr as np_seterr,
                   shape as np_shape,
                   sin as np_sin,
                   size as np_size,
                   sort as np_sort,
                   std as np_std,
                   sum as np_sum,
                   sqrt as np_sqrt,
                   unravel_index as np_unravel_index,
                   where as np_where,
                   zeros as np_zeros)
from numpy.linalg import norm as np_norm
import scipy.ndimage as ndi
from scipy.spatial.distance import pdist, squareform, cityblock, euclidean
from scipy.misc import imsave

# GroopM imports
from profileManager import ProfileManager
from binManager import BinManager
from refine import GrubbsTester, RefineEngine
from PCA import PCA, Center
from groopmExceptions import BinNotFoundException

np_seterr(all='raise')

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class ClusterEngine:
    """Top level interface for clustering contigs"""
    def __init__(self,
                 dbFileName,
                 timer,
                 plot=0,
                 finalPlot=False,
                 force=False,
                 numImgMaps=1,
                 minSize=5,
                 minVol=1000000):
        
        # worker classes
        self.PM = ProfileManager(dbFileName) # store our data
        self.BM = BinManager(pm=self.PM, minSize=minSize, minVol=minVol)

        # heat maps
        self.numImgMaps = numImgMaps
        self.imageMaps = np_zeros((self.numImgMaps,self.PM.scaleFactor,self.PM.scaleFactor))
        self.blurredMaps = np_zeros((self.numImgMaps,self.PM.scaleFactor,self.PM.scaleFactor))

        # we need a way to reference from the imageMaps back onto the transformed data
        self.im2RowIndices = {}

        # When blurring the raw image maps I chose a radius to suit my data, you can vary this as you like
        self.blurRadius = 2
        self.span = 45                  # amount we can travel about when determining "hot spots"

        # Misc tools we'll need
        self.timer = timer
        self.RE = RefineEngine(self.timer, BM=self.BM)   # Basic refinement techniques
        self.HP = HoughPartitioner()                     # Finding cluster centers for unknown K
        self.GT = GrubbsTester()                         # Test length conformity

        # misc
        self.forceWriting = force
        self.debugPlots = plot
        self.finalPlot = finalPlot
        self.imageCounter = 1           # when we print many images
        self.roundNumber = 0            # how many times have we tried to make a bin?
        self.subRoundNumber = 0         # measure sub rounds too!
        self.TSpan = 0.                 # dist from centre to the corner

    def promptOnOverwrite(self, minimal=False):
        """Check that the user is ok with possibly overwriting the DB"""
        if(self.PM.isClustered()):
            if(not self.forceWriting):
                input_not_ok = True
                valid_responses = ['Y','N']
                vrs = ",".join([str.lower(str(x)) for x in valid_responses])
                while(input_not_ok):
                    if(minimal):
                        option = raw_input(" Overwrite? ("+vrs+") : ")
                    else:
                        option = raw_input(" ****WARNING**** Database: '"+self.PM.dbFileName+"' has already been clustered.\n" \
                                           " If you continue you *MAY* overwrite existing bins!\n" \
                                           " Overwrite? ("+vrs+") : ")
                    if(option.upper() in valid_responses):
                        print "****************************************************************"
                        if(option.upper() == "N"):
                            print "Operation cancelled"
                            return False
                        else:
                            break
                    else:
                        print "Error, unrecognised choice '"+option.upper()+"'"
                        minimal = True
            print "Will Overwrite database",self.PM.dbFileName
        return True

#------------------------------------------------------------------------------
# CORE CONSTRUCTION AND MANAGEMENT

    def makeCores(self, coreCut, gf="", kmerThreshold=0.2, coverageThreshold=0.05):
        """Cluster the contigs to make bin cores"""
        # check that the user is OK with nuking stuff...
        if(not self.promptOnOverwrite()):
            return False

        # get some data
        self.PM.loadData(self.timer, condition="length >= "+str(coreCut))
        print "    %s" % self.timer.getTimeStamp()

        # transform the data
        print "    Loading transformed data"
        self.PM.transformCP(self.timer)
        # plot the transformed space (if we've been asked to...)
        if(self.debugPlots >= 3):
            self.PM.renderTransCPData()

        # now we can make this guy
        self.TSpan = np_mean([np_norm(self.PM.corners[i] - self.PM.TCentre) for i in range(self.PM.numStoits)])
        
        print "    %s" % self.timer.getTimeStamp()

        # cluster and bin!
        print "Create cores"
        self.initialiseCores(kmerThreshold, coverageThreshold)
        print "    %s" % self.timer.getTimeStamp()

        # condense cores
        print "Refine cores [begin: %d]" % len(self.BM.bins)
        if self.finalPlot:
            prfx = "CORE"
        else:
            prfx = ""
        self.RE.refineBins(self.timer, auto=True, saveBins=False, plotFinal=prfx, gf=gf)

        # Now save all the stuff to disk!
        print "Saving bins"
        self.BM.saveBins(nuke=True)
        print "    %s" % self.timer.getTimeStamp()

    def initialiseCores(self, kmerThreshold, coverageThreshold):
        """Process contigs and form CORE bins"""
        num_below_cutoff = 0            # how many consecutive attempts have produced small bins
        breakout_point = 30            # how many will we allow before we stop this loop

        # First we need to find the centers of each blob.
        # We can make a heat map and look for hot spots
        self.populateImageMaps()
        sub_counter = 0
        print "     .... .... .... .... .... .... .... .... .... ...."
        print "%4d" % sub_counter,
        new_line_counter = 0
        num_bins = 0

        while(num_below_cutoff < breakout_point):
            stdout.flush()

            # apply a gaussian blur to each image map to make hot spots
            # stand out more from the background
            self.blurMaps()

            # now search for the "hottest" spots on the blurred map
            # and check for possible bin centroids
            putative_clusters = self.findNewClusterCenters(kmerThreshold, coverageThreshold)

            if(putative_clusters is None):
                break
            else:
                bids_made = []
                partitions = putative_clusters[0]
                [max_x, max_y] = putative_clusters[1]
                self.roundNumber += 1
                self.subRoundNumber = 1

                for center_row_indices in partitions:
                    total_BP = np_sum(self.PM.contigLengths[center_row_indices])
                    bin_size = len(center_row_indices)

                    if self.BM.isGoodBin(total_BP, bin_size, ms=3, mv=10000):   # Can we trust very small bins?.
                        # time to make a bin
                        bin = self.BM.makeNewBin(rowIndices=center_row_indices)

                        # work out the distribution in points in this bin
                        bin.makeBinDist(self.PM.transformedCP, self.PM.averageCoverages, self.PM.kmerNormPC1, self.PM.kmerPCs, self.PM.contigGCs, self.PM.contigLengths)

                        # append this bins list of mapped rowIndices to the main list
                        bids_made.append(bin.id)
                        num_bins += 1
                        self.updatePostBin(bin)

                        if(self.debugPlots >= 2):
                            bin.plotBin(self.PM.transformedCP, self.PM.contigGCs, self.PM.kmerNormPC1,
                                        self.PM.contigLengths, self.PM.colorMapGC, self.PM.isLikelyChimeric,
                                        fileName="FRESH_"+str(self.imageCounter))

                            self.imageCounter += 1
                            self.subRoundNumber += 1
                    else:
                        # this partition was too small, restrict these guys we don't run across them again
                        self.restrictRowIndices(center_row_indices)

                # did we do anything?
                num_bids_made = len(bids_made)
                if(num_bids_made == 0):
                    num_below_cutoff += 1
                    # nuke the lot!
                    for row_indices in partitions:
                        self.restrictRowIndices(row_indices)

                # do some post processing
                for bid in bids_made:
                    try:
                        bin = self.BM.getBin(bid)

                        # recruit more contigs
                        bin.recruit(self.PM,
                                    self.GT,
                                    self.im2RowIndices
                                    )
                        self.updatePostBin(bin)

                        new_line_counter += 1
                        print "% 4d" % bin.binSize,

                        # make the printing prettier
                        if(new_line_counter > 9):
                            new_line_counter = 0
                            sub_counter += 10
                            print "\n%4d" % sub_counter,

                        if(self.debugPlots >= 1):
                            #***slow plot!
                            bin.plotBin(self.PM.transformedCP, self.PM.contigGCs, self.PM.kmerNormPC1, self.PM.contigLengths, self.PM.colorMapGC, self.PM.isLikelyChimeric, fileName="CORE_BIN_%d"%(bin.id))

                    except BinNotFoundException: pass

        print "\n     .... .... .... .... .... .... .... .... .... ...."

    def findNewClusterCenters(self, kmerThreshold, coverageThreshold):
        """Find a putative cluster"""
        inRange = lambda x,l,u : x >= l and x < u

        # we work from the top view as this has the base clustering
        max_index = np_argmax(self.blurredMaps[0])
        max_value = self.blurredMaps[0].ravel()[max_index]

        max_x = int(max_index/self.PM.scaleFactor)
        max_y = max_index - self.PM.scaleFactor*max_x

        ret_values = [max_x, max_y]
        
        if(self.debugPlots >= 2):
            self.plotHeat("HM_%d.%d.png" % (self.roundNumber+1, self.subRoundNumber), x=max_x, y=max_y)

        start_span = int(1.5 * self.span)
        span_len = 2*start_span+1

        # make a 3d grid to hold the values
        working_block = np_zeros((span_len, span_len, self.PM.scaleFactor))

        # go through the entire column
        (x_lower, x_upper) = self.makeCoordRanges(max_x, start_span)
        (y_lower, y_upper) = self.makeCoordRanges(max_y, start_span)
        super_putative_row_indices = []
        for p in self.im2RowIndices:
            if inRange(p[0],x_lower,x_upper) and inRange(p[1],y_lower,y_upper):
                for row_index in self.im2RowIndices[p]:
                    # check that the point is real and that it has not yet been binned
                    if row_index not in self.PM.binnedRowIndices and row_index not in self.PM.restrictedRowIndices:
                        # this is an unassigned point.
                        multiplier = np_log10(self.PM.contigLengths[row_index])
                        self.incrementAboutPoint3D(working_block, p[0]-x_lower, p[1]-y_lower, p[2],multiplier=multiplier)
                        super_putative_row_indices.append(row_index)

        # blur and find the highest value
        bwb = ndi.gaussian_filter(working_block, 8)#self.blurRadius)
        densest_index = np_unravel_index(np_argmax(bwb), (np_shape(bwb)))
        max_x = densest_index[0] + x_lower
        max_y = densest_index[1] + y_lower
        max_z = densest_index[2]

        # now get the basic color of this dense point
        putative_center_row_indices = []
        
        # z span should be greater at the corners and shallower at the center
        z_span_min = 100
        z_span_max = 400
        dist_from_centre = euclidean(self.PM.TCentre[:2], np_array([max_x,max_y]))
        z_span = (z_span_max - z_span_min)/self.PM.transRadius*dist_from_centre + z_span_min

        (x_lower, x_upper) = self.makeCoordRanges(max_x, self.span)
        (y_lower, y_upper) = self.makeCoordRanges(max_y, self.span)
        (z_lower, z_upper) = self.makeCoordRanges(max_z, z_span)

        for row_index in super_putative_row_indices:
            p = np_around(self.PM.transformedCP[row_index])
            if inRange(p[0],x_lower,x_upper) and inRange(p[1],y_lower,y_upper) and inRange(p[2],z_lower,z_upper):
                # we are within the range!
                putative_center_row_indices.append(row_index)

        putative_center_row_indices = np_array(putative_center_row_indices)

        # make sure we have something to go on here
        if(np_size(putative_center_row_indices) == 0):
            # it's all over!
            return None

        if(np_size(putative_center_row_indices) == 1):
            # get out of here but keep trying
            # the calling function may restrict these indices
            return [[np_array(putative_center_row_indices)], ret_values]
        else:
            total_BP = np_sum(self.PM.contigLengths[putative_center_row_indices])
            if not self.BM.isGoodBin(total_BP, len(putative_center_row_indices), ms=5): # Can we trust very small bins?.
                # get out of here but keep trying
                # the calling function should restrict these indices
                return [[np_array(putative_center_row_indices)], ret_values]
            else:
                putative_clusters = self.twoWayContraction(putative_center_row_indices, 
                                                           [max_x, max_y],
                                                           kmerThreshold, 
                                                           coverageThreshold)
                if putative_clusters == None:
                    return None

                return [putative_clusters, ret_values]

    def twoWayContraction(self, rowIndices, positionInPlane, kmerThreshold, coverageThreshold):
        """Partition a collection of contigs into 'core' groups"""
        
        # sanity check that there is enough data here to try a determine 'core' groups
        total_BP = np_sum(self.PM.contigLengths[rowIndices])
        if not self.BM.isGoodBin(total_BP, len(rowIndices), ms=5): # Can we trust very small bins?.
            # get out of here but keep trying
            # the calling function should restrict these indices
            return [np_array(rowIndices)]
        
        # make a copy of the data we'll be munging
        k_dat = np_copy(self.PM.kmerPCs[rowIndices])
        #c_dat = np_copy(self.PM.covProfiles[rowIndices])
        c_dat = np_copy(self.PM.transformedCP[rowIndices])
        l_dat = np_copy(self.PM.contigLengths[rowIndices])
        if self.debugPlots >= 2:
            n_dat = np_copy(self.PM.contigNames[rowIndices])
            
        row_indices = np_copy(rowIndices)
        
        # calculate shortest distance to a corner (this isn't currently used)
        #min_dist_to_corner = 1e9
        #for corner in self.PM.corners:
        #    dist = cityblock(corner[:,[0,1]], positionInPlane)
        #   min_dist_to_corner = min(dist, min_dist_to_corner)
        
        # calculate radius threshold in whitened transformed coverage space
        #eps_neighbours = np_max([0.05 * len(rowIndices), np_min([10, len(rowIndices)-1])])
        eps_neighbours = np_min([10, int(len(rowIndices)/2)])
        
        # calculate mean and std in coverage space for whitening data
        c_mean = np_mean(c_dat, axis=0)
        c_std = np_std(c_dat, axis=0)
        c_std += np_where(c_std == 0, 1, 0) # make sure std dev is never zero
        c_whiten_dat = (c_dat-c_mean) / c_std
        
        c_whiten_dist = pdist(c_whiten_dat, 'cityblock')
        c_dist_matrix = squareform(c_whiten_dist)
        c_radius = np_median(np_sort(c_dist_matrix)[:,eps_neighbours-1])
        
        # calculate radius threshold in kmer space
        k_dist = pdist(k_dat, 'cityblock')
        k_dist_matrix = squareform(k_dist)
        k_radius = np_median(np_sort(k_dist_matrix)[:,eps_neighbours-1])
        
        # calculate convergence criteria
        k_converged = kmerThreshold * 30.0 #5e-2 * np_mean(k_dist)
        c_converged = coverageThreshold * 3.4  # 5e-2 * np_mean(c_whiten_dist)
        k_delt = 0.
        c_delt = 0.
        max_iterations = 1  # don't worry about params. Just do once
        
        k_move_perc = 0.1
        c_move_perc = 0.1

        # perform two-way contraction of kmer and coverage space
        iter = 0
        while iter < max_iterations:
            iter += 1
            
            if self.debugPlots >= 2:
                if iter == 1:
                    try:
                        self.cluster_num
                    except:
                        self.cluster_num = 0
                    
                    self.cluster_num += 1
            
                fig = plt.figure()
                ax = fig.add_subplot(111, projection='3d')
                ax.scatter(c_dat[:,0],
                           c_dat[:,1],
                           c_dat[:,2],
                           edgecolors='k',
                           c=self.PM.contigGCs[row_indices],
                           cmap=self.PM.colorMapGC,
                           vmin=0.0, vmax=1.0,
                           s=np_sqrt(l_dat),
                           marker='.')
                
                title = "Points: %s Cd: %f Kd: %f" % (str(len(c_dat[:,0])), c_delt, k_delt)
                plt.title(title)
                
                if iter == 1:
                    xlim = [np_min(c_dat[:,0]) - 0.05*np_min(c_dat[:,0]), np_max(c_dat[:,0]) + 0.05*np_max(c_dat[:,0])]
                    ylim = [np_min(c_dat[:,1]) - 0.05*np_min(c_dat[:,1]), np_max(c_dat[:,1]) + 0.05*np_max(c_dat[:,1])]
                    zlim = [np_min(c_dat[:,2]) - 0.05*np_min(c_dat[:,2]), np_max(c_dat[:,2]) + 0.05*np_max(c_dat[:,2])]
            
                ax.set_xlim(xlim)
                ax.set_ylim(ylim)
                ax.set_zlim(zlim)
                
                fig.set_size_inches(6,6)
                
                fileName = "gh_%d_%d" % (self.cluster_num, iter)
                plt.savefig(fileName + '.png',dpi=96)
                
                plt.close(fig)
                del fig
        
            # calculate distance matrices
            c_dist_matrix = squareform(pdist(c_whiten_dat, 'cityblock'))
            k_dist_matrix = squareform(pdist(k_dat, 'cityblock'))
    
            # find nearest neighbours to each point in whitened coverage space,
            # and use this to converage a point's kmer profile
            new_k_dat = np_zeros(k_dat.shape)
            k_putative_noise = set()
            k_deltas = []
            for index, row in enumerate(c_dist_matrix):
                neigbhours = np_where(row <= c_radius)[0]
                if len(neigbhours) > np_max([1, 0.1*eps_neighbours]):
                    neigbhours = neigbhours[1:] # ignore self match
                else:
                    # extremely few neighbours so mark this as noise
                    k_putative_noise.add(index)
                    new_k_dat[index] = k_dat[index]
                    continue
            
                # use distance between kmer profiles as weights for moving similar
                # points towards each other; a minimum distance based on the kmer
                # radius is used to avoid zeros and ensure all neighbours provide
                # some weight
                neighbour_dist = k_dist_matrix[index][neigbhours] + 0.1 * k_radius
                
                # move point towards neighbours using inverse distance weighting
                inv_dist = 1.0 / neighbour_dist
                sum_inv_dist = np_sum(inv_dist)
                neighbour_weights = inv_dist / sum_inv_dist
                new_k_dat[index] = (1-k_move_perc) * k_dat[index] + k_move_perc * np_sum( (k_dat[neigbhours].T * neighbour_weights).T, axis = 0 )
                
                k_deltas.append(cityblock(k_dat[index], new_k_dat[index]))
            
            k_dat = new_k_dat
            
            # find nearest neighbours to each point in kmer space,
            # and use this to converage a point's coverage profile
            new_c_dat = np_zeros(c_dat.shape)
            c_putative_noise = set()
            c_deltas = []
            for index, row in enumerate(k_dist_matrix):
                neigbhours = np_where(row <= k_radius)[0]
                if len(neigbhours) > np_max([1, 0.1*eps_neighbours]):
                    neigbhours = neigbhours[1:] # ignore self match
                else:
                    # extremely few neighbours so mark this as nois
                    c_putative_noise.add(index)
                    new_c_dat[index] = c_dat[index]
                    continue
                
                # use distance between whitened coverage profiles as weights for moving similar
                # points towards each other; a minimum distance based of the coverage
                # radius is used to avoid zeros and ensure all neighbours provide some weight
                neighbour_dist = c_dist_matrix[index][neigbhours] + 0.1 * c_radius
                
                # move point towards neighbours using inverse distance weighting
                inv_dist = 1.0 / neighbour_dist
                sum_inv_dist = np_sum(inv_dist)
                neighbour_weights = inv_dist / sum_inv_dist
                new_c_dat[index] = (1-c_move_perc) * c_dat[index] + c_move_perc * np_sum( (c_dat[neigbhours].T * neighbour_weights).T, axis = 0 )
                new_c_whiten_dat = (1-c_move_perc) * c_whiten_dat[index] + c_move_perc * np_sum( (c_whiten_dat[neigbhours].T * neighbour_weights).T, axis = 0 )
                
                c_deltas.append(cityblock(c_whiten_dat[index], new_c_whiten_dat))
            
            c_dat = new_c_dat
            
            # remove points that have no or few neighbours in both spaces,
            # unless they are long enough to be of interest
            noise = []
            putative_noise = c_putative_noise.intersection(k_putative_noise)
            for index in putative_noise:
                if not self.BM.isGoodBin(l_dat[index], 1):
                    noise.append(index)
            
            if len(noise) > 0:
                c_dat = np_delete(c_dat, noise, axis = 0)
                k_dat = np_delete(k_dat, noise, axis = 0)
                l_dat = np_delete(l_dat, noise, axis = 0)
                row_indices = np_delete(row_indices, noise, axis = 0)
                if self.debugPlots >= 2:
                    #print "Noise deleting_%d_%d:\n" % (self.cluster_num, iter), n_dat[noise]
                    n_dat = np_delete(n_dat, noise, axis = 0)
            
            # get whitened version of modified coverage data (using original transformation)
            c_whiten_dat = (c_dat-c_mean) / c_std
            
            if len(row_indices) == 0:
                return None
            
            # check for convergence
            k_delt = np_mean(k_deltas)
            c_delt = np_mean(c_deltas)
            if k_delt < k_converged or c_delt < c_converged:
                break

        if self.debugPlots >= 2:
        
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.scatter(c_dat[:,0],
                       c_dat[:,1],
                       c_dat[:,2],
                       edgecolors='k',
                       c=self.PM.contigGCs[row_indices],
                       cmap=self.PM.colorMapGC,
                       vmin=0.0, vmax=1.0,
                       s=np_sqrt(l_dat),
                       marker='.')
            
            title = "Points: %s Cd: %f Kd: %f" % (str(len(c_dat[:,0])), c_delt, k_delt)
            plt.title(title)
            
            if iter == 1:
                xlim = [np_min(c_dat[:,0]) - 0.05*np_min(c_dat[:,0]), np_max(c_dat[:,0]) + 0.05*np_max(c_dat[:,0])]
                ylim = [np_min(c_dat[:,1]) - 0.05*np_min(c_dat[:,1]), np_max(c_dat[:,1]) + 0.05*np_max(c_dat[:,1])]
                zlim = [np_min(c_dat[:,2]) - 0.05*np_min(c_dat[:,2]), np_max(c_dat[:,2]) + 0.05*np_max(c_dat[:,2])]
        
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            ax.set_zlim(zlim)
            
            fig.set_size_inches(6,6)
            
            fileName = "gh_%d_final" % (self.cluster_num)
            plt.savefig(fileName + '.png',dpi=96)
            
            plt.close(fig)
            del fig

        # test total GC spread
        # if it's very small we won't bother doing hough stuff
        gcs = self.PM.contigGCs[row_indices]
        gc_spread = np_max(gcs) - np_min(gcs)
        self.HP.hc += 1
        if gc_spread <= 0.15:
            # no hough
            (k_partitions, k_keeps) = (np_array([np_arange(len(row_indices))]),np_array([True]))
        else:
            # perform hough transform clustering
            if self.debugPlots >= 3:
                (k_partitions, k_keeps) = self.HP.houghPartition(k_dat[:,0], l_dat, imgTag="MER")
            else:
                (k_partitions, k_keeps) = self.HP.houghPartition(k_dat[:,0], l_dat)
        
        if(len(k_partitions) == 0):
            return None
        
        partitions = []
        k_sizes = [len(p) for p in k_partitions]
        
        #-----------------------
        # GRID
        if self.debugPlots >= 2:
            c_max = np_max(c_dat[:,2]/10) * 1.1
            k_max = np_max(k_dat[:,0]) * 1.1
            c_min = np_min(c_dat[:,2]/10) * 0.9
            k_min = np_min(k_dat[:,0]) * 0.9
    
            k_index_sort = np_argsort(k_dat[:,0])
            start = 0
            k_lines = []
            for k in range(len(k_sizes)-1):
                k_lines.append(k_dat[k_index_sort,0][k_sizes[k]+start])
                start += k_sizes[k]
    
            fig = plt.figure()
    
            orig_k_dat = self.PM.kmerPCs[rowIndices,0]
            orig_k2_dat = self.PM.kmerPCs[rowIndices,1]
            orig_c_dat = self.PM.transformedCP[rowIndices][:,2]/10
            orig_l_dat = np_sqrt(self.PM.contigLengths[rowIndices])
    
            ax = plt.subplot(221)
            plt.xlabel("PCA1")
            plt.ylabel("PCA2")
    
            from matplotlib.patches import Rectangle
            alpha = 0.35
            ax.scatter(orig_k_dat, orig_k2_dat, edgecolors='none', c=self.PM.contigGCs[rowIndices], cmap=self.PM.colorMapGC, vmin=0.0, vmax=1.0, s=orig_l_dat, zorder=10, alpha=alpha)
            XX = ax.get_xlim()
            YY = ax.get_ylim()
            ax.add_patch(Rectangle((XX[0], YY[0]),XX[1]-XX[0],YY[1]-YY[0],facecolor='#000000'))
    
            ax = plt.subplot(223)
            plt.title("%s contigs" % len(rowIndices))
            plt.xlabel("MER PARTS")
            plt.ylabel("COV PARTS")
            ax.scatter(orig_k_dat, orig_c_dat, edgecolors='none', c=self.PM.contigGCs[rowIndices], cmap=self.PM.colorMapGC, s=orig_l_dat, zorder=10, alpha=alpha)
            XX = ax.get_xlim()
            YY = ax.get_ylim()
            ax.add_patch(Rectangle((XX[0], YY[0]),XX[1]-XX[0],YY[1]-YY[0],facecolor='#000000'))
    
            lens = np_sqrt(self.PM.contigLengths[row_indices])
    
            ax = plt.subplot(222)
            plt.xlabel("PCA1")
            plt.ylabel("PCA2")
            ax.scatter(k_dat[:,0],
                       k_dat[:,1],
                       edgecolors='none',
                       c=self.PM.contigGCs[row_indices],
                       cmap=self.PM.colorMapGC,
                       vmin=0.0,
                       vmax=1.0,
                       s=lens,
                       zorder=10,
                       alpha=alpha)
            XX = ax.get_xlim()
            YY = ax.get_ylim()
            ax.add_patch(Rectangle((XX[0], YY[0]),XX[1]-XX[0],YY[1]-YY[0],facecolor='#000000'))
    
            ax = plt.subplot(224)
            plt.title("%s contigs" % len(row_indices))
            plt.xlabel("MER PARTS")
            plt.ylabel("COV PARTS")
            ax.scatter(k_dat[:,0],
                       c_dat[:,2]/10,
                       edgecolors='none',
                       c=self.PM.contigGCs[row_indices],
                       cmap=self.PM.colorMapGC,
                       vmin=0.0,
                       vmax=1.0,
                       s=lens,
                       zorder=10,
                       alpha=alpha)
    
            ax.set_xlim(k_min, k_max)
            ax.set_ylim(c_min, c_max)
            XX = ax.get_xlim()
            YY = ax.get_ylim()
            ax.add_patch(Rectangle((XX[0], YY[0]),XX[1]-XX[0],YY[1]-YY[0],facecolor='#000000'))
    
            k_line_cols = []
            for k in range(len(k_sizes)):
                if k == 0:
                    if k_keeps[k]:
                        k_line_cols = ['r-']
                    else:
                        k_line_cols = ['r--']
                elif k == len(k_sizes) - 1:
                    if k_keeps[k]:
                        k_line_cols[-1] = 'r-'
                    elif not k_keeps[k-1]:
                        k_line_cols[-1] = 'r--'
                else:
                    if k_keeps[k]:
                        k_line_cols[k-1] = 'r-'
                        k_line_cols.append('r-')
                    else:
                        k_line_cols.append('r--')
    
            for k in range(len(k_lines)):
                plt.plot([k_lines[k],k_lines[k]], [c_min, c_max], k_line_cols[k], zorder=11)

        pc = 0
        for k in range(len(k_sizes)):
            pc += 1
            if k_keeps[k]:
                k_part = k_partitions[k]
                part_bp = np_sum(l_dat[k_part])
                if self.BM.isGoodBin(part_bp, len(k_part), ms=5):
        
                    # select just the subset of covs for this kmer range
                    data = np_copy(c_dat[k_part])
                    
                    # PCA the subset and cluster on the 1st component
                    Center(data,verbose=0)
                    p = PCA(data)
                    components = p.pc()
                    data = np_array([float(i) for i in components[:,0]])
                    
                    l_data = np_copy(l_dat[k_part])
        
                    # The PCA may reverse the ordering. So we just check here quickly
                    if self.debugPlots >= 3:
                        (c_partitions, c_keeps) = self.HP.houghPartition(data,
                                                                         l_data,
                                                                         imgTag="COV_%s" % (pc-1),
                                                                         gData=self.PM.transformedCP[row_indices[k_part],2],
                                                                         gCut=10.)
                    else:
                        (c_partitions, c_keeps) = self.HP.houghPartition(data,
                                                                         l_data,
                                                                         gData=self.PM.transformedCP[row_indices[k_part],2],
                                                                         gCut=10.)
        
        
                    if self.debugPlots >= 2:
                        #-----
                        # GRID
                        c_sorted_data = np_copy(c_dat[k_part,2])/10.
                        c_sorted_data = c_sorted_data[np_argsort(c_sorted_data)]
        
                        start = 0
                        c_lines = []
                        if k_part[c_partitions[0]][0] > k_part[c_partitions[-1]][0]:
                            c_sizes = [len(p) for p in c_partitions][::-1]
                            cc_keeps = c_keeps[::-1]
                        else:
                            c_sizes = [len(p) for p in c_partitions]
                            cc_keeps = c_keeps
        
                        for c in range(len(c_sizes)-1):
                            c_lines.append(c_sorted_data[c_sizes[c]+start])
                            start += c_sizes[c]
        
                        c_line_cols = []
                        for c in range(len(c_sizes)):
                            if c == 0:
                                if cc_keeps[c]:
                                    c_line_cols = ['r-']
                                else:
                                    c_line_cols = ['r--']
                            elif c == len(c_sizes) - 1:
                                if cc_keeps[c]:
                                    c_line_cols[-1] = 'r-'
                                elif not cc_keeps[c-1]:
                                    c_line_cols[-1] = 'r--'
                            else:
                                if cc_keeps[c]:
                                    c_line_cols[c-1] = 'r-'
                                    c_line_cols.append('r-')
                                else:
                                    c_line_cols.append('r--')
        
                        if pc == 1:
                            k_line_min = k_min
                        else:
                            k_line_min = k_lines[pc-2]
        
                        if pc == len(k_partitions):
                            k_line_max = k_max
                        else:
                            k_line_max = k_lines[pc-1]
        
                        for c in range(len(c_lines)):
                            plt.plot([k_line_min,k_line_max], [c_lines[c], c_lines[c]], c_line_cols[c], zorder=11)
        
                    for c in range(len(c_partitions)):
                        if c_keeps[c]:
                            c_part = c_partitions[c]
                            partitions.append(np_array(k_part[c_part]))

        if self.debugPlots >= 2:
            fig.set_size_inches(12,12)
            plt.savefig("%d_GRID" % self.HP.hc,dpi=300)
            plt.close()
            del fig

        if len(partitions) == 0:
            return None
        
        ret_parts = []
        for p in partitions:
            ret_parts.append(np_array(row_indices[p]))
        
        return np_array(ret_parts)

#------------------------------------------------------------------------------
# DATA MAP MANAGEMENT

    def populateImageMaps(self):
        """Load the transformed data into the main image maps"""
        # reset these guys... JIC
        self.imageMaps = np_zeros((self.numImgMaps,self.PM.scaleFactor,self.PM.scaleFactor))
        self.im2RowIndices = {}

        # add to the grid wherever we find a contig
        row_index = -1
        for point in np_around(self.PM.transformedCP):
            row_index += 1
            # can only bin things once!
            if row_index not in self.PM.binnedRowIndices and row_index not in self.PM.restrictedRowIndices:
                # add to the row_index dict so we can relate the
                # map back to individual points later
                p = tuple(point)
                try:
                    self.im2RowIndices[p].append(row_index)
                except KeyError:
                    self.im2RowIndices[p] = [row_index]

                # now increment in the grid
                # for each point we encounter we incrmement
                # it's position + the positions to each side
                # and touching each corner
                self.incrementViaRowIndex(row_index, p)

    def incrementViaRowIndex(self, rowIndex, point=None):
        """Wrapper to increment about point"""
        if(point is None):
            point = tuple(np_around(self.PM.transformedCP[rowIndex]))
        multiplier = np_log10(self.PM.contigLengths[rowIndex])
        self.incrementAboutPoint(0, point[0], point[1], multiplier=multiplier)
        if(self.numImgMaps > 1):
            self.incrementAboutPoint(1, self.PM.scaleFactor - point[2] - 1, point[1], multiplier=multiplier)
            self.incrementAboutPoint(2, self.PM.scaleFactor - point[2] - 1, self.PM.scaleFactor - point[0] - 1, multiplier=multiplier)

    def decrementViaRowIndex(self, rowIndex, point=None):
        """Wrapper to decrement about point"""
        if(point is None):
            point = tuple(np_around(self.PM.transformedCP[rowIndex]))
        #px = point[0]
        #py = point[1]
        #pz = point[2]
        multiplier = np_log10(self.PM.contigLengths[rowIndex])
        self.decrementAboutPoint(0, point[0], point[1], multiplier=multiplier)
        if(self.numImgMaps > 1):
            self.decrementAboutPoint(1, self.PM.scaleFactor - point[2] - 1, point[1], multiplier=multiplier)
            self.decrementAboutPoint(2, self.PM.scaleFactor - point[2] - 1, self.PM.scaleFactor - point[0] - 1, multiplier=multiplier)

    def incrementAboutPoint(self, view_index, px, py, valP=1, valS=0.6, valC=0.2, multiplier=1):
        """Increment value at a point in the 2D image maps

        Increment point by valP, increment neighbouring points at the
        sides and corners of the target point by valS and valC

        multiplier is proportional to the contigs length
        """
        valP *= multiplier
        valS *= multiplier
        valC *= multiplier
        if px > 0:
            if py > 0:
                self.imageMaps[view_index,px-1,py-1] += valC      # Top left corner
            self.imageMaps[view_index,px-1,py] += valS            # Top
            if py < self.PM.scaleFactor-1:
                self.imageMaps[view_index,px-1,py+1] += valC      # Top right corner

        if py > 0:
            self.imageMaps[view_index,px,py-1] += valS            # Left side
        self.imageMaps[view_index,px,py] += valP                  # Point
        if py < self.PM.scaleFactor-1:
            self.imageMaps[view_index,px,py+1] += valS            # Right side

        if px < self.PM.scaleFactor-1:
            if py > 0:
                self.imageMaps[view_index,px+1,py-1] += valC      # Bottom left corner
            self.imageMaps[view_index,px+1,py] += valS            # Bottom
            if py < self.PM.scaleFactor-1:
                self.imageMaps[view_index,px+1,py+1] += valC      # Bottom right corner

    def decrementAboutPoint(self, view_index, px, py, valP=1, valS=0.6, valC=0.2, multiplier=1):
        """Decrement value at a point in the 2D image maps

        multiplier is proportional to the contigs length
        """
        valP *= multiplier
        valS *= multiplier
        valC *= multiplier
        if px > 0:
            if py > 0:
                self.safeDecrement(self.imageMaps[view_index], px-1, py-1, valC) # Top left corner
            self.safeDecrement(self.imageMaps[view_index], px-1, py, valS)       # Top
            if py < self.PM.scaleFactor-1:
                self.safeDecrement(self.imageMaps[view_index], px-1, py+1, valC) # Top right corner

        if py > 0:
            self.safeDecrement(self.imageMaps[view_index], px, py-1, valS)       # Left side
        self.safeDecrement(self.imageMaps[view_index], px, py, valP)             # Point
        if py < self.PM.scaleFactor-1:
            self.safeDecrement(self.imageMaps[view_index], px, py+1, valS)       # Right side

        if px < self.PM.scaleFactor-1:
            if py > 0:
                self.safeDecrement(self.imageMaps[view_index], px+1, py-1, valC) # Bottom left corner
            self.safeDecrement(self.imageMaps[view_index], px+1, py, valS)       # Bottom
            if py < self.PM.scaleFactor-1:
                self.safeDecrement(self.imageMaps[view_index], px+1, py+1, valC) # Bottom right corner

    def safeDecrement(self, map, px, py, value):
        """Decrement a value and make sure it's not negative or something shitty"""
        map[px][py] -= value
        if map[px][py] < np_finfo(float).eps:
            map[px][py] = 0

    def incrementAboutPoint3D(self, workingBlock, px, py, pz, vals=(6.4,4.9,2.5,1.6), multiplier=1):
        """Increment a point found in a 3D column

        used when finding the centroid of a hot area
        update the 26 points which surround the centre point
        z spans the height of the entire column, x and y have been offset to
        match the column subspace

        multiplier is proportional to the contigs length
        """
        valsM = [x*multiplier for x in vals]
        # top slice
        if pz < self.PM.scaleFactor-1:
            self.subIncrement3D(workingBlock, px, py, pz+1, valsM, 1)

        # center slice
        self.subIncrement3D(workingBlock, px, py, pz, valsM, 0)

        # bottom slice
        if pz > 0:
            self.subIncrement3D(workingBlock, px, py, pz-1, valsM, 1)

    def subIncrement3D(self, workingBlock, px, py, pz, vals, offset):
        """AUX: Called from incrementAboutPoint3D does but one slice

        multiplier is proportional to the contigs length
        """
        # get the size of the working block
        shape = np_shape(workingBlock)
        if px > 0:
            if py > 0:
                workingBlock[px-1,py-1,pz] += vals[offset + 2]      # Top left corner
            workingBlock[px-1,py,pz] += vals[offset + 1]            # Top
            if py < shape[1]-1:
                workingBlock[px-1,py+1,pz] += vals[offset + 2]      # Top right corner

        if py > 0:
            workingBlock[px,py-1,pz] += vals[offset + 1]            # Left side
        workingBlock[px,py,pz] += vals[offset]                      # Point
        if py < shape[1]-1:
            workingBlock[px,py+1,pz] += vals[offset + 1]            # Right side

        if px < shape[0]-1:
            if py > 0:
                workingBlock[px+1,py-1,pz] += vals[offset + 2]      # Bottom left corner
            workingBlock[px+1,py,pz] += vals[offset + 1]            # Bottom
            if py < shape[1]-1:
                workingBlock[px+1,py+1,pz] += vals[offset + 2]      # Bottom right corner

    def blurMaps(self):
        """Blur the 2D image maps"""
        self.blurredMaps = np_zeros((self.numImgMaps,self.PM.scaleFactor,self.PM.scaleFactor))
        for i in range(self.numImgMaps): # top, front and side
            self.blurredMaps[i,:,:] = ndi.gaussian_filter(self.imageMaps[i,:,:], 8)#self.blurRadius)

    def makeCoordRanges(self, pos, span):
        """Make search ranges which won't go out of bounds"""
        lower = pos-span
        upper = pos+span+1
        if(lower < 0):
            lower = 0
        if(upper > self.PM.scaleFactor):
            upper = self.PM.scaleFactor
        return (lower, upper)

    def updatePostBin(self, bin):
        """Update data structures after assigning contigs to a new bin"""
        bid = bin.id
        for row_index in bin.rowIndices:
            self.PM.binIds[row_index] = bid
            self.setRowIndexAssigned(row_index)

    def setRowIndexAssigned(self, rowIndex):
        """fix the data structures to indicate that rowIndex belongs to a bin

        Use only during initial core creation
        """
        if(rowIndex not in self.PM.restrictedRowIndices and rowIndex not in self.PM.binnedRowIndices):
            self.PM.binnedRowIndices[rowIndex] = True
            # now update the image map, decrement
            self.decrementViaRowIndex(rowIndex)

    def setRowIndexUnassigned(self, rowIndex):
        """fix the data structures to indicate that rowIndex no longer belongs to a bin

        Use only during initial core creation
        """
        if(rowIndex in self.PM.restrictedRowIndices and rowIndex not in self.PM.binnedRowIndices):
            del self.PM.binnedRowIndices[rowIndex]
            # now update the image map, increment
            self.incrementViaRowIndex(rowIndex)

    def restrictRowIndices(self, indices):
        """Add these indices to the restricted list"""
        for row_index in indices:
            # check that it's not binned or already restricted
            if(row_index not in self.PM.restrictedRowIndices and row_index not in self.PM.binnedRowIndices):
                self.PM.restrictedRowIndices[row_index] = True
                # now update the image map, decrement
                self.decrementViaRowIndex(row_index)

#------------------------------------------------------------------------------
# IO and IMAGE RENDERING

    def plotIndices(self, rowIndices, fileName=""):
        """Plot these contigs in transformed space"""
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        disp_vals = np_array([])
        disp_lens = np_array([])
        num_points = 0
        for row_index in rowIndices:
            num_points += 1
            disp_vals = np_append(disp_vals, self.PM.transformedCP[row_index])
            disp_lens = np_append(disp_lens, np_sqrt(self.PM.contigLengths[row_index]))

        # reshape
        disp_vals = np_reshape(disp_vals, (num_points, 3))

        ax.scatter(disp_vals[:,0], disp_vals[:,1], disp_vals[:,2], edgecolors='k', c=self.PM.contigGCs[rowIndices], cmap=self.PM.colorMapGC, vmin=0.0, vmax=1.0, s=disp_lens, marker='.')

        if(fileName != ""):
            fig.set_size_inches(6,6)
            plt.savefig(fileName,dpi=300)
        elif(show):
            plt.show()

        plt.close(fig)
        del fig

    def plotRegion(self, px, py, pz, fileName="", tag="", column=False):
        """Plot the region surrounding a point """
        import matplotlib as mpl
        disp_vals = np_array([])
        disp_cols = np_array([])
        num_points = 0
        # plot all points within span
        (z_lower, z_upper) = self.makeCoordRanges(pz, self.span)
        if(column):
            z_lower = 0
            z_upper = self.PM.scaleFactor - 1

        (x_lower, x_upper) = self.makeCoordRanges(px, self.span)
        (y_lower, y_upper) = self.makeCoordRanges(py, self.span)
        for z in range(z_lower, z_upper):
            realz = self.PM.scaleFactor - z - 1
            for x in range(x_lower, x_upper):
                for y in range(y_lower, y_upper):
                    if((x,y,realz) in self.im2RowIndices):
                        for row_index in self.im2RowIndices[(x,y,realz)]:
                            if row_index not in self.PM.binnedRowIndices and row_index not in self.PM.restrictedRowIndices:
                                num_points += 1
                                disp_vals = np_append(disp_vals, self.PM.transformedCP[row_index])
                                disp_cols = np_append(disp_cols, self.PM.colorMapGC(self.PM.contigGCs[row_index]))

        # make a black mark at the max values
        small_span = self.span/2
        (x_lower, x_upper) = self.makeCoordRanges(px, small_span)
        (y_lower, y_upper) = self.makeCoordRanges(py, small_span)
        (z_lower, z_upper) = self.makeCoordRanges(pz, small_span)
        for z in range(z_lower, z_upper):
            realz = self.PM.scaleFactor - z - 1
            for x in range(x_lower, x_upper):
                for y in range(y_lower, y_upper):
                    if((x,y,realz) in self.im2RowIndices):
                        for row_index in self.im2RowIndices[(x,y,realz)]:
                            if row_index not in self.PM.binnedRowIndices and row_index not in self.PM.restrictedRowIndices:
                                num_points += 1
                                disp_vals = np_append(disp_vals, self.PM.transformedCP[row_index])
                                disp_cols = np_append(disp_cols, htr(0,0,0))
        # reshape
        disp_vals = np_reshape(disp_vals, (num_points, 3))
        disp_cols = np_reshape(disp_cols, (num_points, 3))

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        cm = mpl.colors.LinearSegmentedColormap('my_colormap', disp_cols, 1024)
        sc = ax.scatter(disp_vals[:,0], disp_vals[:,1], disp_vals[:,2], edgecolors=disp_cols, c=disp_cols, cmap=cm, marker='.')
        sc.set_edgecolors = sc.set_facecolors = lambda *args:None # disable depth transparency effect
        title = str.join(" ", ["Focus at: (",str(px), str(py), str(self.PM.scaleFactor - pz - 1),")\n",tag])
        plt.title(title)

        if(fileName != ""):
            fig.set_size_inches(6,6)
            plt.savefig(fileName,dpi=300)
        elif(show):
            plt.show()

        plt.close(fig)
        del fig

    def plotHeat(self, fileName = "", x=-1, y=-1):
        """Print the main heat maps

        Useful for debugging
        """
        fig = plt.figure()
        images = []
        ax = None
        if(self.numImgMaps == 1):
            ax = fig.add_subplot(121)
            ax.imshow(self.blurredMaps[0,:,:]**0.5)
            if x != -1:
                ax.scatter([y],
                           [x],
                           edgecolors='#ff69b4',
                           c='none',
                           s=5000.,
                           marker='.')
        else:
            ax = fig.add_subplot(231)
            images.append(ax.imshow(self.blurredMaps[0,:,:]**0.5))
            ax = fig.add_subplot(232)
            images.append(ax.imshow(self.blurredMaps[1,:,:]**0.5))
            ax = fig.add_subplot(233)
            images.append(ax.imshow(self.blurredMaps[2,:,:]**0.5))

        if(self.numImgMaps == 1):
            ax = fig.add_subplot(122)
            ax.imshow(self.imageMaps[0,:,:]**0.5)
        else:
            ax = fig.add_subplot(234)
            images.append(ax.imshow(self.imageMaps[0,:,:]**0.5))
            ax = fig.add_subplot(235)
            images.append(ax.imshow(self.imageMaps[1,:,:]**0.5))
            ax = fig.add_subplot(236)
            images.append(ax.imshow(self.imageMaps[2,:,:]**0.5))

        if(fileName != ""):
            if(self.numImgMaps == 1):
                fig.set_size_inches(12,6)
            else:
                fig.set_size_inches(18,18)

            plt.savefig(fileName,dpi=300)
        elif(show):
            plt.show()

        plt.close(fig)
        del fig

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class HoughPartitioner:
    def __init__(self):
        self.hc = 0

    def houghPartition(self,
                       dAta,            # data to cluster with
                       lData,           # lengths for filling in about long contigs
                       imgTag=None,     # make debug plots
                       gData=None,      # gap data to check partitions with
                       gCut=0           # gap cutoff to stop merging of partitions
                       ):
        """Use the hough transform to find k clusters for some unknown value of K"""
        d_len_raw = int(len(dAta))
        if d_len_raw < 2:
            return np_array([[0]])

        if d_len_raw < 3:
            return np_array([[0,1]])

        #----------------------------------------------------------------------
        # prep the data
        #
        sorted_indices_raw = np_argsort(dAta)
        nUm_points = len(dAta)

        # fudge the data to make longer contigs have more say in the
        # diff line we'll be making. This way we may be able to avoid lumping
        # super long contigs in with the short riff raff by accident.
        scales_per = {}
        spread_data = []
        spread2real = {}
        j = 0

        # all points get at least one point, but long ones get more
        # let's say 1 point per 2000bp
        for i in range(len(dAta)):
            real_index = sorted_indices_raw[i]
            rep = int((lData[real_index] - 1.)/5000.) + 1
            scales_per[real_index] = rep
            if rep == 1:
                spread_data.append(dAta[real_index])
                spread2real[j] = real_index
                j += 1
            else:
                # rep >= 2
                if i == 0:
                    left_stop = dAta[real_index]
                else:
                    left_stop = (dAta[real_index] + dAta[sorted_indices_raw[i-1]])/2.

                if i == nUm_points-1:
                    right_stop = dAta[real_index]
                else:
                    right_stop = (dAta[real_index] + dAta[sorted_indices_raw[i+1]])/2.

                spread_jump = (right_stop - left_stop) / (rep + 1.)


                for ii in range(rep):
                    spread_data.append(left_stop + ((ii + 1) * spread_jump))
                    spread2real[j] = real_index
                    j += 1

        # Force the data to fill the space
        data = np_array(spread_data)
        data -= np_min(data)    # shift to 0 but DO NOT scale to 1
        d_len = len(data)
        scale = np_max(dAta) - np_min(dAta)

        # we want to know how much each value differs from it's neighbours
        back_diffs = [float(data[i] - data[i-1]) for i in range(1,d_len)]
        diffs = [back_diffs[0]]
        for i in range(len(back_diffs)-1):
            diffs.append((back_diffs[i] + back_diffs[i+1])/2)
        diffs.append(back_diffs[-1])
        diffs = np_array(diffs)**2  # square it! Makes things more betterrer

        # replace the data array by the sum of it's diffs
        for i in range(1, d_len):
            diffs[i] += diffs[i-1]

        # scale to fit between 0 and len(diffs)
        # HT works better on a square
        diffs -= np_min(diffs)
        try:
            diffs /= np_max(diffs)
        except FloatingPointError:
            pass
        diffs *= len(diffs)

        # make it 2D
        t_data = np_array(zip(diffs, np_arange(d_len)))
        im_shape = (int(np_max(t_data, axis=0)[0]+1), d_len)

        #----------------------------------------------------------------------
        # Apply hough transformation and find the most prominent line
        #
        # rets should be an array of arrays of real indices
        # IE. indices into the array dAta
        rets = self.recursiveSelect(t_data,
                                    im_shape,
                                    spread2real,
                                    0,
                                    d_len,
                                    {},
                                    imgTag=imgTag)

        #----------------------------------------------------------------------
        # Squish things up
        #
        # build a flat data set similar to the gradiated data set
        real2spread = {}
        j = 0
        flat_data = []
        if len(rets) > 1:
            for i in range(len(dAta)):
                real_index = sorted_indices_raw[i]
                rep = scales_per[real_index]
                for k in range(rep):
                    flat_data.append(dAta[real_index])
                    try:
                        real2spread[real_index].append(j)
                    except KeyError:
                        real2spread[real_index] = [j]
                    j += 1
            data -= np_min(flat_data)
            back_diffs = [float(data[i] - data[i-1]) for i in range(1,d_len)]
            diffs = [back_diffs[0]]
            for i in range(len(back_diffs)-1):
                diffs.append((back_diffs[i] + back_diffs[i+1])/2)
            diffs.append(back_diffs[-1])
            diffs = np_array(diffs)**2
            for i in range(1, d_len):
                diffs[i] += diffs[i-1]
            diffs -= np_min(diffs)
            try:
                diffs /= np_max(diffs)
            except FloatingPointError:
                pass
            diffs *= len(diffs)

            # diffs is now the same size as the gradiated data sent through
            # to hough in level 0. We wish to find the gradients of the lines
            # returned by recursive partitioning
            gradients = []
            for ret in rets:
                sis = []
                for ii in ret:
                    for si in real2spread[ii]:
                        sis.append(diffs[si])
                l_sis = len(sis)
                if l_sis == 1:
                    gradients.append(-1)
                else:
                    sis = sorted(sis)
                    gradients.append((sis[-1] - sis[0])/l_sis)

            gradients = np_array(gradients)

            # get all the -1 gradients and make them equal to the larger
            # of their neighbours
            last = 0.
            for g in range(len(gradients)):
                if gradients[g] == -1:
                    h = g + 1
                    next = 0.
                    # find the next not -1 gradient
                    while(h < len(gradients)):
                        if gradients[h] != -1:
                            # use this one
                            next = gradients[h]
                            break
                        h += 1
                    gradients[g] = np_max([last, next])
                else:
                    last = gradients[g]

        else:
            gradients=np_array([0.])

        keeps = np_where(gradients >= 1, False, True)

        squished_rets = []
        squished_keeps = []
        last_squished = []
        
        if gData is not None:
            # measure the gaps between potential squishables
            last_max = -1
            gaps = []
            for i in range(len(rets)):
                tmp_gs = []
                for ii in rets[i]:
                    tmp_gs.append(gData[ii])    # collate the gData for this partition
                    
                A = np_min(tmp_gs)              # find it's boundaries
                B = np_max(tmp_gs)
                if last_max > 0:
                    gaps.append(A - last_max)   # measure the gap
                last_max = B
            gap_info = [True] + [ggg < gCut for ggg in gaps]
        else:
            gap_info = [True] * len(keeps)
            
        for i in range(len(rets)):
            if keeps[i]:
                # check and see if we are allowed to squish
                if gap_info[i]: 
                    # squish OK
                    for ii in rets[i]:
                        last_squished.append(ii)
                else:
                    # shouldn't squish
                    # push on the last fella
                    if len(last_squished) > 0:
                        squished_rets.append(np_array(last_squished))
                        squished_keeps.append(True)
                    # remake the last_squished holder
                    last_squished = []
                    for ii in rets[i]:
                        last_squished.append(ii)
            else:
                if len(last_squished) > 0:
                    squished_rets.append(np_array(last_squished))
                    last_squished = []
                    squished_keeps.append(True)

                # add the dud
                squished_rets.append(rets[i])
                squished_keeps.append(False)
        if len(last_squished) > 0:
            squished_rets.append(np_array(last_squished))
            squished_keeps.append(True)

        return (np_array(squished_rets), np_array(squished_keeps))

    def recursiveSelect(self,
                        tData,
                        imShape,
                        spread2real,
                        startRange,
                        endRange,
                        assigned,
                        level=0,
                        side="C",
                        imgTag=None):
        """Recursively select clusters from the data"""
        d_len = len(tData)
        (m, ret_point, accumulator) = self.houghTransform(tData.astype(float)[startRange:endRange,:], imShape)

        if m == np_inf:
            # this is a vertical line through ret_point
            start_p = [0, ret_point[1]]
            end_p = [imShape[0], ret_point[1]]
        else:
            # find out where the line crosses axes
            y_int = ret_point[0] - m * ret_point[1]
            if m == 0:
                # line is flat
                x_int = np_inf
            else:
                x_int = ret_point[1] - ret_point[0] / m

            if np_abs(y_int) < np_abs(x_int):
                # m line passes above (0,0)
                start_p = [y_int, 0]
                end_p = [imShape[1]*m + y_int, imShape[1]]
            else:
                # m line passes below (0,0)
                start_p = [0, x_int]
                end_p = [imShape[0], imShape[0]/m + x_int]

        # draw a nice thick line over the top of the data
        # found_line is a set of points
        found_line = self.points2Line(np_array([start_p,end_p]), imShape[1], imShape[0], 5)

        # make an image if we're that way inclined
        if imgTag is not None:
            # make a pretty picture
            fff = np_ones(imShape) * 255
            for p in found_line.keys():
                fff[p[0],p[1]] = 220
            for p in tData:
                fff[p[0],p[1]] = 0
            
            # scale so colors look sharper
            accumulator -= np_min(accumulator)
            accumulator /= np_max(accumulator)
            accumulator *= 255

            imsave("%d_%s_%s_%d.png" % (self.hc, imgTag, side, level), np_concatenate([accumulator,fff]))

        # see which points lie on the line
        # we need to protect against the data line crossing
        # in and out of the "found line"
        in_block = False
        block_starts = []
        block_lens = []
        for ii in np_arange(startRange, endRange):
            p = tData.astype('int')[ii]
            if tuple(p) in found_line:
                if not in_block:
                    in_block = True
                    block_starts.append(ii)
            else:
                if in_block:
                    in_block = False
                    block_lens.append(ii - block_starts[-1])

        if in_block:
            # finishing block
            block_lens.append(endRange - block_starts[-1])

        # check to see the line hit something
        if len(block_lens) == 0:
            tmp = {}
            for ii in np_arange(startRange, endRange):
                real_index = spread2real[ii]
                if real_index not in assigned:
                    tmp[real_index] = None
                    assigned[real_index] = None
            centre = np_array(tmp.keys())
            if len(centre) > 0:
                return np_array([centre])
            # nuffin
            return np_array([])

        # get the start and end indices in the longest block found
        longest_block = np_argmax(block_lens)
        spread_start = block_starts[longest_block]
        spread_end =  block_lens[longest_block] + spread_start  # 1 after the end of the block

        # select all the guys with their centres between the start and end
        # this is the end of the line for these guys so we fill centre with
        # "real" indices.
        tmp = {}
        for ii in np_arange(spread_start, spread_end):
            real_index = spread2real[ii]
            if real_index not in assigned:
                tmp[real_index] = None
                assigned[real_index] = None
        centre = np_array(tmp.keys())

        rets = []

        # recursive call for leftmost indices
        if (spread_start - startRange) > 0:
            if (spread_start - startRange) < 3:
                # end of the line for left expansion, give up "real" indices
                # select all the guys with their centres to the left of the start
                tmp = {}
                for ii in np_arange(startRange, spread_start):
                    real_index = spread2real[ii]
                    if real_index not in assigned:
                        tmp[real_index] = None
                        assigned[real_index] = None

                if len(tmp.keys()) > 0:
                    rets.append(np_array(tmp.keys()))

            else:
                # otherwise we keep working with ranges
                left_p = self.recursiveSelect(tData,
                                              imShape,
                                              spread2real,
                                              startRange,
                                              spread_start,
                                              assigned,
                                              level=level+1,
                                              side="%sL" %side,
                                              imgTag=imgTag)
                for L in left_p:
                    if len(L) > 0:
                        rets.append(L)

        # add the centre in
        if len(centre) > 0:
            rets.append(centre)

        # recursive call for rightmost indices
        if (endRange - spread_end) > 0:
            if (endRange - spread_end) < 3:
                # end of the line for left expansion, give up "real" indices
                # select all the guys with their centres right of the end
                tmp = {}
                for ii in np_arange(spread_end, endRange):
                    real_index = spread2real[ii]
                    if real_index not in assigned:
                        tmp[real_index] = None
                        assigned[real_index] = None

                if len(tmp.keys()) > 0:
                    rets.append(np_array(tmp.keys()))
            else:
                right_p = self.recursiveSelect(tData,
                                               imShape,
                                               spread2real,
                                               spread_end,
                                               endRange,
                                               assigned,
                                               level=level+1,
                                               side="%sR" %side,
                                               imgTag=imgTag)
                for R in right_p:
                    if len(R) > 0:
                        rets.append(R)
        return np_array(rets)

    def points2Line(self, points, xIndexLim, yIndexLim, thickness):
        """Draw a thick line between a series of points"""

        line_points = []
        num_points = len(points)
        for i in range(1, num_points):
            # draw a line between this point and the last point
            x_gap = float(np_abs(points[i-1,1] - points[i,1]))
            y_gap = float(np_abs(points[i-1,0] - points[i,0]))
            largest_gap = np_max([x_gap, y_gap])

            if points[i-1,0] >= points[i,0]:
                Ys = [int(j) for j in np_around((np_arange(largest_gap)*y_gap/largest_gap) + points[i,0])]
            elif points[i,0] > points[i-1,0]:
                Ys = [int(j) for j in np_around((np_arange(largest_gap)*y_gap/largest_gap) + points[i-1,0])][::-1]
            if points[i-1,1] >= points[i,1]:
                Xs = [int(j) for j in np_around((np_arange(largest_gap)*x_gap/largest_gap) + points[i,1])]
            elif points[i,1] > points[i-1,1]:
                Xs = [int(j) for j in np_around((np_arange(largest_gap)*x_gap/largest_gap) + points[i-1,1])][::-1]

            for p in zip(Ys, Xs):
                line_points.append(p)

            # now make the line thicker
            thick_points = {}
            for point in line_points:
                for y in range(np_max([point[0]-thickness, 0]),np_min([point[0]+thickness+1,yIndexLim])):
                    for x in range(np_max([point[1]-thickness, 0]),np_min([point[1]+thickness+1,xIndexLim])):
                        thick_points[(y,x)] = 1

        return thick_points

    def houghTransform(self, data, imShape):
        """Calculate Hough transform

        Data is a 2D numpy array"""

        (rows, cols) = imShape
        d_len = len(data)
        half_rows = rows/2
        if half_rows == 0:
            half_rows = 1
            rows = 2
        rmax = np_hypot(rows, cols)
        dr = rmax / (half_rows)
        dth = np_pi / cols
        accumulator = np_ones((rows * cols))*255

        """
        For speed we numpify this loop. I just keep this here
        so that I can remember what it is I am actually doing...

        Note that this needs the accumulator to be a 2D array
        ie.
        accumulator = np_ones((rows, cols))*255

        for p in data:
            for theta_index in range(cols):
                th = dth * theta_index
                r = p[1]*cos(th) + p[0]*sin(th)
                iry = half_rows + int(r/dr)
                accumulator[iry, theta_index] -= 1
        """
        cos_sin_array = np_array(zip([np_sin(dth * theta_index) for theta_index in range(cols)],
                                     [np_cos(dth * theta_index) for theta_index in range(cols)]))
        Rs = np_array(np_sum(np_reshape([p * cos_sin_array for p in data], (d_len*cols,2)),
                             axis=1)/dr).astype('int') + half_rows
        Cs = np_array(range(cols)*d_len)
        flat_indices = Rs * cols + Cs

        # update the accumulator with integer decrements
        decrements = np_bincount(flat_indices.astype('int'))
        index = 0
        for d in decrements:
            accumulator[index] -= d
            index += 1

        minindex = accumulator.argmin()

        # find the liniest line
        min_row = int(minindex/cols)
        min_col = minindex - (min_row*cols)
        theta = float(min_col) * dth
        rad = float(min_row - half_rows)*dr

        # now de hough!
        # first get a point our found line passes through
        Ps = []
        for p in range(len(data)):
            point = data[p]
            th = dth * min_col
            r = point[1]*np_cos(th) + point[0]*np_sin(th)
            iry = half_rows + int(r/dr)
            # this point does
            if iry == min_row:
                Ps.append(data[p])
        # take the average of all of em'
        ret_point = np_mean(Ps, axis=0)
        
        # get the gradient
        if theta != 0:
            m = -1. * np_cos(theta) / np_sin(theta)
        else:
            m = np_inf

        # rounding errors suck
        if np_allclose([m], [0.]):
            m = 0.

        return (m, ret_point, accumulator.reshape((rows, cols)))

###############################################################################
###############################################################################
###############################################################################
###############################################################################



from distutils.core import setup

setup(
    name='GroopM',
    version='0.2.10.1',
    author='Michael Imelfort',
    author_email='mike@mikeimelfort.com',
    packages=['groopm', 'groopm.test'],
    scripts=['bin/groopm'],
	url='http://pypi.python.org/pypi/GroopM/',
    license='LICENSE.txt',
    description='Metagenomic binning suite',
    long_description=open('README.txt').read(),
    install_requires=[
        "numpy >= 1.6.1",
        "scipy >= 0.10.1",
        "matplotlib >= 1.1.0",
        "tables >= 2.3",
        "pysam >= 0.6",
        "PIL >= 1.1.7",
        "BamTyper >= 0.2.6"
    ],
)



#!/usr/bin/python
###############################################################################
#                                                                             #
#    ellipsoid.py                                                             #
#                                                                             #
#    Playing with ellipses!                                                   #
#                                                                             #
#    Copyright (C) Michael Imelfort                                           #
#                                                                             #
###############################################################################
#                                                                             #
#          .d8888b.                                    888b     d888          #
#         d88P  Y88b                                   8888b   d8888          #
#         888    888                                   88888b.d88888          #
#         888        888d888 .d88b.   .d88b.  88888b.  888Y88888P888          #
#         888  88888 888P"  d88""88b d88""88b 888 "88b 888 Y888P 888          #
#         888    888 888    888  888 888  888 888  888 888  Y8P  888          #
#         Y88b  d88P 888    Y88..88P Y88..88P 888 d88P 888   "   888          #
#          "Y8888P88 888     "Y88P"   "Y88P"  88888P"  888       888          #
#                                             888                             #
#                                             888                             #
#                                             888                             #
#                                                                             #
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Michael Imelfort"
__copyright__ = "Copyright 2012/2013"
__credits__ = ["Michael Imelfort"]
__license__ = "GPL3"
__version__ = "0.1.1"
__maintainer__ = "Michael Imelfort"
__email__ = "mike@mikeimelfort.com"
__status__ = "Beta"

###############################################################################

import matplotlib.pyplot as plt
import numpy as np
from numpy import linalg

np.seterr(all='raise')

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class EllipsoidTool:
    """Some stuff for playing with ellipsoids"""
    def __init__(self): pass

    def getMinVolEllipse(self, P, tolerance=0.01, retA=False):
        """ Find the minimum volume ellipsoid which holds all the points

        Based on work by Nima Moshtagh
        http://www.mathworks.com/matlabcentral/fileexchange/9542
        and also by looking at:
        http://cctbx.sourceforge.net/current/python/scitbx.math.minimum_covering_ellipsoid.html
        Which is based on the first reference anyway!

        Here, P is a numpy array of 3D points like this:
        P = [[x,y,z],
             [x,y,z],
             [x,y,z]]

        Returns:
        (center, radii, rotation)

        """
        (N, d) = np.shape(P)

        # Q will be out working array
        Q = np.copy(P.T)
        Q = np.vstack([Q, np.ones(N)])
        QT = Q.T

        # initializations
        err = 1 + tolerance
        u = np.array([1.0 / N for i in range(N)]) # first iteration

        # Khachiyan Algorithm
        singular = False
        while err > tolerance:
            V = np.dot(Q, np.dot(np.diag(u), QT))
            try:
                M = np.diag(np.dot(QT , np.dot(linalg.inv(V), Q)))    # M the diagonal vector of an NxN matrix
            except linalg.linalg.LinAlgError:
                # most likely a singular matrix
                # permute the values a little and then we'll try again
                from random import random, randint
                PP = np.copy(P)
                for i in range(N):
                    if randint(0,3) == 0:
                        j = randint(0,2)
                        if randint(0,1) != 0:
                            PP[i,j] += random()
                        else:
                            PP[i,j] -= random()
                (A, center, radii, rotation) = self.getMinVolEllipse(PP, retA=True)
                singular = True
                break

            j = np.argmax(M)
            maximum = M[j]
            step_size = (maximum - d - 1.0) / ((d + 1.0) * (maximum - 1.0))
            new_u = (1.0 - step_size) * u
            new_u[j] += step_size
            err = np.linalg.norm(new_u - u)
            u = new_u

        if not singular:
            # center of the ellipse
            center = np.dot(P.T, u)

            # the A matrix for the ellipse
            try:
                A = linalg.inv(
                               np.dot(P.T, np.dot(np.diag(u), P)) -
                               np.array([[a * b for b in center] for a in center])
                               ) / d
            except linalg.linalg.LinAlgError:
                # the matrix is singular so we need to return a degenerate ellipse
                #print '[Notice] Degenerate ellipse constructed indicating a bin with extremely small coverage divergence.'
                center = np.mean(P, axis=0)
                radii = np.max(P,axis=0) - np.min(P, axis=0)

                if len(P[0]) == 3:
                    rotation = [[0,0,0],[0,0,0],[0,0,0]]
                else:
                    rotation = [[0,0],[0,0]]

                if retA:
                    return (None, center, radii, rotation)
                else:
                    return (center, radii, rotation)

        # Get the values we'd like to return
        try:
            U, s, rotation = linalg.svd(A)
            radii = 1.0/np.sqrt(s)
        except np.linalg.linalg.LinAlgError:
            # hack -> better than crashing...
            rotation = np.eye(3)
            radii = np.ones(3)
        if retA:
            return (A, center, radii, rotation)
        else:
            return (center, radii, rotation)

    def getEllipsoidVolume(self, radii):
        """Calculate the volume of the blob"""
        if len(radii) == 2:
            return np.pi*radii[0]*radii[1]
        else:
            return (4.0/3.0)*np.pi*radii[0]*radii[1]*radii[2]

    def doesIntersect3D(self, A, cA, B, cB):
        """Rough test to see if ellipsoids A and B intersect

        Not perfect, should work for "well overlapping" ones
        We assume that the volume of B is less than (or =) volume of A
        """
        #To make things simple, we just check if the points on a wire frame of
        #B lie within A

        # Quick check if the centre of B is within ellipse A. This deals with
        # degenerate cases where B is only a single point or an otherwise
        # degenerate ellipse.
        p_c = cB - cA
        if np.dot(p_c.T, np.dot(A, p_c)) <= 1:
            return True

        if A == None or B == None: # degenerate ellipse that can't be processed
            return False

        U, s, rotation = linalg.svd(B)
        try:
            radii_B = 1.0/np.sqrt(s)
        except FloatingPointError:
            # the given matrix B was made on a group of only one point
            # we need only check if the one point (the center)
            # in in A
            p_c = cB - cA
            return np.dot(p_c.T, np.dot(A, p_c)) <= 1

        u = np.linspace(0.0, 2.0 * np.pi, 100)
        v = np.linspace(0.0, np.pi, 100)

        # cartesian coordinates that correspond to the spherical angles:
        x = radii_B[0] * np.outer(np.cos(u), np.sin(v))
        y = radii_B[1] * np.outer(np.sin(u), np.sin(v))
        z = radii_B[2] * np.outer(np.ones_like(u), np.cos(v))

        # rotate accordingly
        for i in range(len(x)):
            for j in range(len(x)):
                # make a point on the wireFrame
                wire_point = np.dot([x[i,j],y[i,j],z[i,j]], rotation) + cB
                # test if it's inside
                # work out (p-c)'A(p-c) and see if it's <= 1
                p_c = wire_point - cA
                if np.dot(p_c.T, np.dot(A, p_c)) <= 1:
                    return True

        return False

    def doesIntersect2D(self, A, cA, B, cB):
        """Rough test to see if ellipsoids A and B intersect

        Not perfect, should work for "well overlapping" ones
        We assume that the volume of B is less than (or =) volume of A
        """
        #To make things simple, we just check if the points on a wire frame of
        #B lie within A

        # Quick check if the centre of B is within ellipse A. This deals with
        # degenerate cases where B is only a single point or an otherwise
        # degenerate ellipse.
        p_c = cB - cA
        if np.dot(p_c.T, np.dot(A, p_c)) <= 1:
            return True

        if A == None or B == None:  # degenerate ellipse that can't be processed
            return False

        U, s, rotation = linalg.svd(B)
        try:
            radii_B = 1.0/np.sqrt(s)
        except FloatingPointError:
            # the given matrix B was made on a group of only one point
            # we need only check if the one point (the center)
            # in in A
            p_c = cB - cA
            return np.dot(p_c.T, np.dot(A, p_c)) <= 1

        u = np.linspace(0.0, 2.0 * np.pi, 100)

        # cartesian coordinates that correspond to the spherical angles:
        x = radii_B[0] * np.cos(u)
        y = radii_B[1] * np.sin(u)
        # rotate accordingly
        for i in range(len(x)):
            # make a point on the wireFrame
            edge_point = np.dot([x[i],y[i]], rotation) + cB
            # test if it's inside
            # work out (p-c)'A(p-c) and see if it's <= 1
            p_c = edge_point - cA
            if np.dot(p_c.T, np.dot(A, p_c)) <= 1:
                return True
        return False

    def plotEllipsoid(self, center, radii, rotation, ax=None, plotAxes=False, cageColor='b', cageAlpha=0.2, label=None):
        """Plot an ellipsoid"""
        make_ax = ax == None
        if make_ax:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')

        u = np.linspace(0.0, 2.0 * np.pi, 100)
        v = np.linspace(0.0, np.pi, 100)

        # cartesian coordinates that correspond to the spherical angles:
        x = radii[0] * np.outer(np.cos(u), np.sin(v))
        y = radii[1] * np.outer(np.sin(u), np.sin(v))
        z = radii[2] * np.outer(np.ones_like(u), np.cos(v))
        # rotate accordingly
        for i in range(len(x)):
            for j in range(len(x)):
                [x[i,j],y[i,j],z[i,j]] = np.dot([x[i,j],y[i,j],z[i,j]], rotation) + center

        if plotAxes:
            # make some purdy axes
            axes = np.array([[radii[0],0.0,0.0],
                             [0.0,radii[1],0.0],
                             [0.0,0.0,radii[2]]])
            # rotate accordingly
            for i in range(len(axes)):
                axes[i] = np.dot(axes[i], rotation)


            # plot axes
            for p in axes:
                X3 = np.linspace(-p[0], p[0], 100) + center[0]
                Y3 = np.linspace(-p[1], p[1], 100) + center[1]
                Z3 = np.linspace(-p[2], p[2], 100) + center[2]
                ax.plot(X3, Y3, Z3, color=cageColor)

        # plot ellipsoid
        ax.plot_wireframe(x, y, z,  rstride=4, cstride=4, color=cageColor, alpha=cageAlpha)

        if label is not None:
            ax.text(center[0],
                    center[1],
                    center[2],
                    label,
                    color=[0,0,0],
                    weight='bold'
                    )

        if make_ax:
            plt.show()
            plt.close(fig)
            del fig

    def plotEllipse(self, center, radii, rotation, ax=None, plotAxes=False, cageColor='b', cageAlpha=0.2, label=None, linewidth=-1):
        """plot an ellipse"""
        make_ax = ax == None
        if make_ax:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        u = np.linspace(0.0, 2.0 * np.pi, 100)

        # cartesian coordinates that correspond to the spherical angles:
        x = radii[0] * np.cos(u)
        y = radii[1] * np.sin(u)

        # rotate accordingly
        for i in range(len(x)):
            [x[i],y[i]] = np.dot([x[i],y[i]], rotation) + center

        if plotAxes:
            # make some purdy axes
            axes = np.array([[radii[0],0.0],[0.0,radii[1]]])
            # rotate accordingly
            for i in range(len(axes)):
                axes[i] = np.dot(axes[i], rotation)

            # plot axes
            for p in axes:
                X3 = np.linspace(-p[0], p[0], 100) + center[0]
                Y3 = np.linspace(-p[1], p[1], 100) + center[1]
                ax.plot(X3, Y3, color=cageColor)

        # plot ellipsoid
        if linewidth == -1:
            ax.plot(x, y, color=cageColor, alpha=cageAlpha)
        else:
            ax.plot(x, y, color=cageColor, alpha=cageAlpha, linewidth=linewidth, zorder = 10)

        if label is not None:
            ax.text(center[0],
                    center[1],
                    label,
                    color=[0,0,0],
                    weight='bold'
                    )

        if make_ax:
            plt.show()
            plt.close(fig)
            del fig

###############################################################################
###############################################################################
###############################################################################
###############################################################################



#!/usr/bin/env python
###############################################################################
#                                                                             #
#    binManager.py                                                            #
#                                                                             #
#    GroopM - High level bin data management                                  #
#                                                                             #
#    Copyright (C) Michael Imelfort                                           #
#                                                                             #
###############################################################################
#                                                                             #
#          .d8888b.                                    888b     d888          #
#         d88P  Y88b                                   8888b   d8888          #
#         888    888                                   88888b.d88888          #
#         888        888d888 .d88b.   .d88b.  88888b.  888Y88888P888          #
#         888  88888 888P"  d88""88b d88""88b 888 "88b 888 Y888P 888          #
#         888    888 888    888  888 888  888 888  888 888  Y8P  888          #
#         Y88b  d88P 888    Y88..88P Y88..88P 888 d88P 888   "   888          #
#          "Y8888P88 888     "Y88P"   "Y88P"  88888P"  888       888          #
#                                             888                             #
#                                             888                             #
#                                             888                             #
#                                                                             #
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Michael Imelfort"
__copyright__ = "Copyright 2012/2013"
__credits__ = ["Michael Imelfort"]
__license__ = "GPL3"
__version__ = "0.2.4"
__maintainer__ = "Michael Imelfort"
__email__ = "mike@mikeimelfort.com"
__status__ = "Beta"

###############################################################################
from os.path import join as osp_join
from sys import exc_info, exit, stdout as sys_stdout
from operator import itemgetter


import matplotlib.pyplot as plt
from pylab import show
from numpy import (abs as np_abs,
                   append as np_append,
                   arange as np_arange,
                   arccos as np_arccos,
                   argmax as np_argmax,
                   argsort as np_argsort,
                   around as np_around,
                   array as np_array,
                   bincount as np_bincount,
                   ceil as np_ceil,
                   concatenate as np_concatenate,
                   dot as np_dot,
                   log10 as np_log10,
                   max as np_max,
                   mean as np_mean,
                   median as np_median,
                   min as np_min,
                   ones as np_ones,
                   reshape as np_reshape,
                   seterr as np_seterr,
                   size as np_size,
                   sort as np_sort,
                   sqrt as np_sqrt,
                   std as np_std,
                   sum as np_sum,
                   where as np_where,
                   zeros as np_zeros)

from scipy.stats import f_oneway, distributions
from scipy.cluster.vq import kmeans,vq

# GroopM imports
from profileManager import ProfileManager
from bin import Bin
import groopmExceptions as ge
from groopmUtils import makeSurePathExists
from ellipsoid import EllipsoidTool

np_seterr(all='raise')

###############################################################################
###############################################################################
###############################################################################
###############################################################################

class BinManager:
    """Class used for manipulating bins"""
    def __init__(self,
                 dbFileName="",
                 pm=None,
                 minSize=10,
                 minVol=1000000):
        # data storage
        if(dbFileName != ""):
            self.PM = ProfileManager(dbFileName)
        elif(pm is not None):
            self.PM = pm

        # all about bins
        self.nextFreeBinId = 0                      # increment before use!
        self.bins = {}                              # bid -> Bin

        # misc
        self.minSize=minSize           # Min number of contigs for a bin to be considered legit
        self.minVol=minVol             # Override on the min size, if we have this many BP

    def setColorMap(self, colorMapStr):
        self.PM.setColorMap(colorMapStr)

#------------------------------------------------------------------------------
# LOADING / SAVING

    def loadBins(self,
                 timer,
                 getUnbinned=False,
                 bids=[],
                 makeBins=False,
                 silent=True,
                 loadKmerPCs=False,
                 loadRawKmers=False,
                 loadCovProfiles=True,
                 loadContigLengths=True,
                 loadLinks=False,
                 loadContigNames=True,
                 cutOff=0,
                 transform=True):
        """Load data and make bin objects"""
        # fix the condition
        condition=""
        if(cutOff != 0):
            condition="length >= %d" % cutOff
        elif(len(bids) == 0):
            if getUnbinned:
                condition="length >= 0"
            else:
                condition='bid != 0'

        # if we're going to make bins then we'll need kmer sigs
        if(makeBins):
            loadKmerPCs=True
            loadCovProfiles=True

        self.PM.loadData(timer,
                         bids=bids,
                         condition=condition,
                         silent=silent,
                         loadCovProfiles=loadCovProfiles,
                         loadKmerPCs=loadKmerPCs,
                         loadRawKmers=loadRawKmers,
                         makeColors=True,
                         loadContigNames=loadContigNames,
                         loadContigLengths=loadContigLengths,
                         loadBins=True,
                         loadLinks=loadLinks
                        )

        # exit if no bins loaded
        if self.PM.numContigs == 0:
            return

        if(makeBins):
            if transform:
                self.PM.transformCP(timer, silent=silent)
            else:
                if self.PM.numStoits == 3:
                    self.PM.transformedCP = self.PM.covProfiles
                else:
                    print "Number of stoits != 3. You need to transform"
                    self.PM.transformCP(timer, silent=silent)
            if not silent:
                print "    Making bin objects"
            self.makeBins(self.getBinMembers())
            if not silent:
                print "    Loaded %d bins from database" % len(self.bins)
        if not silent:
            print "    %s" % timer.getTimeStamp()
            sys_stdout.flush()

    def getBinMembers(self):
        """Munge the raw data into something more usable

        self.PM.binIds is an array, contains 0 for unassigned rows
        By default this creates an array for the '0' bin. You need
        to ignore it later if you want to
        """
        # fill them up
        bin_members = {0:[]}
        for row_index in range(np_size(self.PM.indices)):
            try:
                bin_members[self.PM.binIds[row_index]].append(row_index)
            except KeyError:
                bin_members[self.PM.binIds[row_index]] = [row_index]

        # we need to get the largest BinId in use
        if len(bin_members) > 0:
            self.nextFreeBinId = np_max(bin_members.keys())
        return bin_members

    def makeBins(self, binMembers, zeroIsBin=False):
        """Make bin objects from loaded data"""
        invalid_bids = []
        for bid in binMembers:
            if bid != 0 or zeroIsBin:
                if len(binMembers[bid]) == 0:
                    invalid_bids.append(bid)
                else:
                    self.bins[bid] = Bin(np_array(binMembers[bid]), bid, self.PM.scaleFactor-1)
                    self.bins[bid].makeBinDist(self.PM.transformedCP, self.PM.averageCoverages, self.PM.kmerNormPC1, self.PM.kmerPCs, self.PM.contigGCs, self.PM.contigLengths)
        if len(invalid_bids) != 0:
            print "MT bins!"
            print invalid_bids
            exit(-1)

    def saveBins(self, binAssignments={}, nuke=False):
        """Save binning results

        binAssignments is a hash of LOCAL row indices Vs bin ids
        { row_index : bid }
        PM.setBinAssignments needs GLOBAL row indices

        We always overwrite the bins table (It is smallish)
        """
        # save the bin assignments
        self.PM.setBinAssignments(
                                  self.getGlobalBinAssignments(binAssignments), # convert to global indices
                                  nuke=nuke
                                  )
        # overwrite the bins table
        self.setBinStats()

    def getGlobalBinAssignments(self, binAssignments={}):
        """Merge the bids, raw DB indexes and core information so we can save to disk

        returns a hash of type:

        { global_index : bid }
        """
        # we need a mapping from cid (or local index) to to global index to binID
        bin_assignment_update = {}

        if binAssignments != {}:
            # we have been told to do only these guys
            for row_index in binAssignments:
                bin_assignment_update[self.PM.indices[row_index]] = binAssignments[row_index]

        else:
            # this are all our regularly binned guys
            for bid in self.getBids():
                for row_index in self.bins[bid].rowIndices:
                    bin_assignment_update[self.PM.indices[row_index]] = bid

        return bin_assignment_update

    def setBinStats(self):
        """Update / overwrite the table holding the bin stats

        Note that this call effectively nukes the existing table
        """

        # create and array of tuples:
        # [(bid, size, likelyChimeric)]
        bin_stats = []
        for bid in self.getBids():
            # no point in saving empty bins
            if np_size(self.bins[bid].rowIndices) > 0:
                bin_stats.append((bid, np_size(self.bins[bid].rowIndices), self.PM.isLikelyChimeric[bid]))
        self.PM.setBinStats(bin_stats)


#------------------------------------------------------------------------------
# REMOVE ALREADY LOADED BINS

    def removeBinAndIndices(self, bid):
        """Remove indices from the PM based on bin identity

        "unload" some data
        """
        # get some info
        rem_bin = self.getBin(bid)
        original_length = len(self.PM.indices)
        rem_list = np_sort(rem_bin.rowIndices)

        # affect the raw data in the PM
        self.PM.reduceIndices(rem_list)
        del self.PM.validBinIds[bid]

        # remove the bin here
        del self.bins[bid]

        # now fix all the rowIndices in all the other bins
        for bid in self.getBids():
            self.bins[bid].rowIndices = self.fixRowIndexLists(original_length, np_sort(self.bins[bid].rowIndices), rem_list)

    def fixRowIndexLists(self, originalLength, oldList, remList):
        """Fix up row index lists which reference into the
        data structure after a call to reduceIndices

        originalLength is the length of all possible row indices
        before the removal (ie self.indices)
        oldList is the old list of row indices
        remList is the list of indices to be removed

        BOTH OLD AND REM LIST MUST BE SORTED ASCENDING!
        """
        shift_down = 0;
        old_list_index = 0
        new_list = np_array([])
        for i in range(originalLength):
            if(i in remList):
                shift_down+=1
            elif(i in oldList):
                new_list = np_append(new_list, oldList[old_list_index]-shift_down)
                old_list_index += 1
        return new_list


#------------------------------------------------------------------------------
# LINKS

    def getLinkingContigs(self, bid):
        """Get all contigs and their bin IDs which link to contigs in this bin"""
        bin2count = {}
        for row_index in bin.rowIndices:
            try:
                #print row_index, len(self.PM.links[row_index]), self.PM.links[row_index], "===",
                for link in self.PM.links[row_index]:
                    #print "{{{",link,"}}}",
                    try:
                        link_bid = self.PM.binIds[link[0]]
                        #print ";;", link_bid, bid ,
                        if link_bid != bid and link_bid != 0:
                            try:
                                bin2count[link_bid] += 1.0
                            except KeyError:
                                bin2count[link_bid] = 1.0
                    except KeyError:
                        pass#print "****\n\n"
                #print "[[[[\n\n"
            except KeyError:
                pass
        #print bin2count
        return bin2count

    def getConnectedBins(self, rowIndex):
        """Get a  list of bins connected to this contig"""
        ret_links = []
        for link in self.PM.links[rowIndex]:
            cid = link[0]
            try:
                bid = self.PM.binIds[cid]
            except KeyError:
                bid = 0
            ret_links.append((cid, bid, link[1]))
        return ret_links

    def getAllLinks(self):
        """Return a sorted array of all links between all bins"""
        bids = self.getBids()
        # first, work out who links with whom...
        all_links = {}
        for bid in bids:
            links = self.getLinkingContigs(bid)
            # links is a hash of type bid : num_links
            for link in links:
                key = self.makeBidKey(bid, link)
                if key not in all_links:
                    all_links[key] = links[link]

        # sort and return
        return sorted(all_links.iteritems(), key=itemgetter(1), reverse=True)

    def getWithinLinkProfiles(self):
        """Determine the average number of links between contigs for all bins"""
        bids = self.getBids()
        link_profiles = {}
        for bid in bids:
            link_profiles[bid] = self.getWithinBinLinkProfile(bid)
        return link_profiles

    def getWithinBinLinkProfile(self, bid):
        """Determine the average number of links between contigs in a bin"""
        links = []
        min_links = 1000000000
        for row_index in bin.rowIndices:
            try:
                for link in self.PM.links[row_index]:
                    link_bid = self.PM.binIds[link[0]]
                    if link_bid == bid:
                        links.append(link[1])
                        if link[1] < min_links:
                            min_links = link[1]
            except KeyError:
                pass
        return (np_mean(links), np_std(links), min_links)

#------------------------------------------------------------------------------
# BIN UTILITIES

    def isGoodBin(self, totalBP, binSize, ms=0, mv=0):
        """Does this bin meet my exacting requirements?"""
        if(ms == 0):
            ms = self.minSize               # let the user choose
        if(mv == 0):
            mv = self.minVol                # let the user choose

        if(totalBP < mv):                   # less than the good volume
            if(binSize > ms):               # but has enough contigs
                return True
        else:                               # contains enough bp to pass regardless of number of contigs
            return True
        return False

    def getBids(self):
        """Return a sorted list of bin ids"""
        return sorted(self.bins.keys())

    def getCentroidProfiles(self, mode="mer"):
        """Return an array containing the centroid stats for each bin"""
        if(mode == "mer"):
            ret_vecs = np_zeros((len(self.bins)))
            outer_index = 0
            for bid in self.getBids():
                ret_vecs[outer_index] = self.bins[bid].kValMeanNormPC1
                outer_index += 1
            return ret_vecs
        elif(mode == "cov"):
            ret_vecs = np_zeros((len(self.bins), len(self.PM.transformedCP[0])))
            outer_index = 0
            for bid in self.getBids():
                ret_vecs[outer_index] = self.bins[bid].covMedians
                outer_index += 1
            return ret_vecs
        else:
            raise ge.ModeNotAppropriateException("Mode",mode,"unknown")

    def split(self, bid, n, mode='kmer', auto=False, saveBins=False, verbose=False, printInstructions=True, use_elipses=True):
        """split a bin into n parts

        if auto == True, then just railroad the split
        if test == True, then test via merging
        if savebins == True, save the split (if you will do it)
        if MCut != 0, carry the split through only if both daughter bins have an M
          less than MCut
        """
        # we need to work out which profile to cluster on
        if(printInstructions and not auto):
            self.printSplitInstructions()

        # make some split bins
        # bids[0] holds the original bin id
        (bin_assignment_update, bids) = self.getSplitties(bid, n, mode)

        if(auto and saveBins):
            print 'here!!!!'
            # charge on through
            self.deleteBins([bids[0]], force=True)  # delete the combined bin
            # save new bins
            self.saveBins(binAssignments=bin_assignment_update)
            return

        # we will need to confer with the user
        # plot some stuff
        # sort the bins by kmer val
        bid_tuples = [(tbid, self.bins[tbid].kValMeanNormPC1) for tbid in bids[1:]]
        bid_tuples.sort(key=itemgetter(1))
        index = 1
        for pair in bid_tuples:
            bids[index] = pair[0]
            index += 1

        self.plotSideBySide(bids, use_elipses=use_elipses)

        user_option = self.promptOnSplit(n,mode)
        if(user_option == 'Y'):
            if(saveBins):
                # save the temp bins
                self.deleteBins([bids[0]], force=True)  # delete the combined bin
                # save new bins
                self.saveBins(binAssignments=bin_assignment_update)
            return

        # If we're here than we don't need the temp bins
        # remove this query from the list so we don't delete him
        del bids[0]
        self.deleteBins(bids, force=True)

        # see what the user wants to do
        if(user_option == 'N'):
            return
        elif(user_option == 'C'):
            self.split(bid,
                       n,
                       mode='cov',
                       auto=auto,
                       saveBins=saveBins,
                       verbose=verbose,
                       printInstructions=False,
                       use_elipses=use_elipses)
        elif(user_option == 'K'):
            self.split(bid,
                       n,
                       mode='kmer',
                       auto=auto,
                       saveBins=saveBins,
                       verbose=verbose,
                       printInstructions=False,
                       use_elipses=use_elipses)

        elif(user_option == 'P'):
            not_got_parts = True
            parts = 0
            while(not_got_parts):
                try:
                    parts = int(raw_input("Enter new number of parts:"))
                except ValueError:
                    print "You need to enter an integer value!"
                    parts = 0
                if(1 == parts):
                    print "Don't be a silly sausage!"
                elif(0 != parts):
                    not_got_parts = False
                    self.split(bid,
                               parts,
                               mode=mode,
                               auto=auto,
                               saveBins=saveBins,
                               verbose=verbose,
                               printInstructions=False,
                               use_elipses=use_elipses)


    def getSplitties(self, bid, n, mode):
        """Return a set of split bins"""
        obs = np_array([])
        if(mode=='kmer'):
            obs = np_array([self.PM.kmerNormPC1[i] for i in self.getBin(bid).rowIndices])
        elif(mode=='cov'):
            obs = np_array([self.PM.covProfiles[i] for i in self.getBin(bid).rowIndices])

        # do the clustering
        try:
            centroids,_ = kmeans(obs,n)
        except ValueError:
            return False
        idx,_ = vq(obs,centroids)

        # build some temp bins
        # this way we can show the user what the split will look like
        idx_sorted = np_argsort(np_array(idx))
        current_group = 0
        bids = [bid]
        bin_assignment_update = {} # row index to bin id
        holding_array = np_array([])
        split_bin = None
        for i in idx_sorted:
            if(idx[i] != current_group):
                # bin is full!
                holding_array = holding_array.astype(int)
                split_bin = self.makeNewBin(holding_array)

                for row_index in holding_array:
                    bin_assignment_update[row_index] = split_bin.id
                split_bin.makeBinDist(self.PM.transformedCP, self.PM.averageCoverages, self.PM.kmerNormPC1, self.PM.kmerPCs, self.PM.contigGCs, self.PM.contigLengths)
                bids.append(split_bin.id)
                holding_array = np_array([])
                current_group = idx[i]
            holding_array = np_append(holding_array, int(self.getBin(bid).rowIndices[i]))

        # do the last one
        if(np_size(holding_array) != 0):
            holding_array = holding_array.astype(int)
            split_bin = self.makeNewBin(holding_array)
            for row_index in holding_array:
                bin_assignment_update[int(row_index)] = split_bin.id
            split_bin.makeBinDist(self.PM.transformedCP, self.PM.averageCoverages, self.PM.kmerNormPC1, self.PM.kmerPCs, self.PM.contigGCs, self.PM.contigLengths)
            bids.append(split_bin.id)

        return (bin_assignment_update, bids)

    def shouldMerge(self, bin1, bin2, ignoreCov=False, ignoreMer=False, merTol=0, confidence=0.95, verbose=False):
        """Determine whether its wise to merge two bins

        Perfoms a one-way anova to determine if the larger bin would be
        significantly changed if it consumed the smaller

        OR does a tolerance test on kmerNormPC1. We assume that bin1 is larger than bin2
        """
        if(bin1.id != bin2.id):
            if not ignoreCov: # work out coverage distributions
                b1_c_dist = bin1.getAverageCoverageDist(self.PM.averageCoverages)
                b2_c_dist = bin2.getAverageCoverageDist(self.PM.averageCoverages)
                c_dist_1 = b1_c_dist
                if(bin1.binSize < bin2.binSize):
                    c_dist_1 = b2_c_dist
                c_dist_2 = np_append(b2_c_dist, b1_c_dist)
                if verbose:
                    tag = "COV:"
                else:
                    tag = ""
                cov_match = self.isSameVariance(c_dist_1, c_dist_2, confidence=confidence, tag=tag)
            else:
                cov_match = True

            if not ignoreMer: # work out kmer distributions
                if not cov_match:
                    return False
                if merTol != 0:
                    # Tolerance based testing
                    upper_k_val_cut = bin1.kValMeanNormPC1 + merTol * bin1.kValStdevNormPC1
                    lower_k_val_cut = bin1.kValMeanNormPC1 - merTol * bin1.kValStdevNormPC1

                    if bin2.kValMeanNormPC1 >= lower_k_val_cut and bin2.kValMeanNormPC1 <= upper_k_val_cut:
                        mer_match = True
                    else:
                        mer_match = False
                else:
                    b1_k_dist = bin1.getkmerValDist(self.PM.kmerNormPC1)
                    b2_k_dist = bin2.getkmerValDist(self.PM.kmerNormPC1)
                    k_dist_1 = b1_k_dist
                    if(bin1.binSize < bin2.binSize):
                        k_dist_1 = b2_k_dist
                    k_dist_2 = np_append(b2_k_dist, b1_k_dist)
                    if verbose:
                        tag = "MER: %0.4f %0.4f" % (np_mean(k_dist_2), np_std(k_dist_2))
                    else:
                        tag = ""
                    mer_match = self.isSameVariance(k_dist_1, k_dist_2, confidence=confidence, tag=tag)
            else:
                mer_match = True

            return cov_match and mer_match

        return False

    def isSameVariance(self, dist1, dist2, confidence=0.95, tag=""):
        """Test to see if the kmerValues for two bins are the same"""
        F_cutoff =  distributions.f.ppf(confidence, 2, len(dist1)+len(dist2)-2)
        F_value = f_oneway(dist1,dist2)[0]
        if tag != "":
            print "%s [V: %f, C: %f]" % (tag, F_value, F_cutoff)
        return F_value < F_cutoff

    def merge(self, bids, auto=False, manual=False, newBid=False, saveBins=False, verbose=False, printInstructions=True, use_elipses=True):
        """Merge two or more bins

        It's a bit strange to have both manual and auto params
        NOTE: manual ALWAYS overrides auto. In the condensing code, auto is
        set programmaticaly, manual is always set by the user. So we listen
        to manual first
        """
        parent_bin = None

        if(printInstructions and not auto):
            self.printMergeInstructions()

        if(newBid):
            # we need to make this into a new bin
            parent_bin = self.makeNewBin()
            
            # now merge it with the first in the new list
            dead_bin = self.getBin(bids[0])
            for row_index in dead_bin.rowIndices:
                self.PM.binIds[row_index] = parent_bin.id
            parent_bin.consume(self.PM.transformedCP, self.PM.averageCoverages, self.PM.kmerNormPC1, self.PM.kmerPCs, self.PM.contigGCs, self.PM.contigLengths, dead_bin, verbose=verbose)
            self.deleteBins([bids[0]], force=True)
        else:
            # just use the first given as the parent
            parent_bin = self.getBin(bids[0])
            
        # a merged bin specified by the user should not be considered chimeric since 
        # they are indicating both original bins were reasonable
        if manual:
            self.PM.isLikelyChimeric[parent_bin.id] = False

        # let this guy consume all the other guys
        ret_val = 0
        some_merged = False
        for i in range(1,len(bids)):
            continue_merge = False
            dead_bin = self.getBin(bids[i])
            if(auto and not manual):
                ret_val = 2
                continue_merge = True
            else:
                tmp_bin = self.makeNewBin(np_concatenate([parent_bin.rowIndices,dead_bin.rowIndices]))
                tmp_bin.makeBinDist(self.PM.transformedCP, self.PM.averageCoverages, self.PM.kmerNormPC1, self.PM.kmerPCs, self.PM.contigGCs, self.PM.contigLengths)
                self.plotSideBySide([parent_bin.id,dead_bin.id,tmp_bin.id], use_elipses=use_elipses)
                self.deleteBins([tmp_bin.id], force=True)
                user_option = self.promptOnMerge(bids=[parent_bin.id,dead_bin.id])
                if(user_option == "N"):
                    print "Merge skipped"
                    ret_val = 1
                    continue_merge=False
                elif(user_option == "Q"):
                    print "All mergers skipped"
                    return 0
                else:
                    ret_val = 2
                    continue_merge=True
            if(continue_merge):

                for row_index in dead_bin.rowIndices:
                    self.PM.binIds[row_index] = parent_bin.id

                parent_bin.consume(self.PM.transformedCP,
                                   self.PM.averageCoverages,
                                   self.PM.kmerNormPC1,
                                   self.PM.kmerPCs,
                                   self.PM.contigGCs,
                                   self.PM.contigLengths,
                                   dead_bin,
                                   verbose=verbose)
                self.deleteBins([bids[i]], force=True)
                some_merged = True

        if some_merged:
            # Fix up the r2b indices and bin updates
            parent_bid = parent_bin.id
            bin_assignment_update = {}
            for row_index in parent_bin.rowIndices:
                bin_assignment_update[row_index] = parent_bid
                try:
                    self.PM.binIds[row_index] = parent_bid
                except KeyError:
                    pass

            if saveBins:
                self.saveBins(binAssignments=bin_assignment_update)

        return ret_val

    def makeBidKey(self, bid1, bid2):
        """Make a unique key from two bids"""
        if(bid1 < bid2):
            return (bid1, bid2)
        return (bid2, bid1)

    def getBin(self, bid):
        """get a bin or raise an error"""
        if bid in self.bins:
            return self.bins[bid]
        else:
            raise ge.BinNotFoundException("Cannot find: "+str(bid)+" in bins dicts")

    def getChimericBinIds(self):
        bids = []
        for bid in self.bins:
            if self.PM.isLikelyChimeric[bid]:
                bids.append(bid)

        return bids

    def getNonChimericBinIds(self):
        bids = []
        for bid in self.bins:
            if not self.PM.isLikelyChimeric[bid]:
                bids.append(bid)
        
        return bids

    def deleteBins(self, bids, force=False, freeBinnedRowIndices=False, saveBins=False):
        """Purge a bin from our lists"""
        if(not force):
            user_option = self.promptOnDelete(bids)
            if(user_option != 'Y'):
                return False
        bin_assignment_update = {}
        for bid in bids:
            if bid in self.bins:
                if(freeBinnedRowIndices):
                    for row_index in self.bins[bid].rowIndices:
                        try:
                            del self.PM.binnedRowIndices[row_index]
                        except KeyError:
                            print bid, row_index, "FUNG"
                        self.PM.binIds[row_index] = 0

                        bin_assignment_update[row_index] = 0
                del self.bins[bid]
                del self.PM.isLikelyChimeric[bid]
            else:
                raise ge.BinNotFoundException("Cannot find: "+str(bid)+" in bins dicts")

        if(saveBins):
            self.saveBins(binAssignments=bin_assignment_update)
        return True

    def makeNewBin(self, rowIndices=np_array([]), bid=None):
        """Make a new bin and add to the list of existing bins"""
        if bid is None:
            self.nextFreeBinId +=1
            bid = self.nextFreeBinId

        self.PM.isLikelyChimeric[bid] = False
        self.bins[bid] = Bin(rowIndices, bid, self.PM.scaleFactor-1)
        return self.bins[bid]

#------------------------------------------------------------------------------
# UI

    def printMergeInstructions(self):
        raw_input( "****************************************************************\n"
                   " MERGING INSTRUCTIONS - PLEASE READ CAREFULLY\n"
                   "****************************************************************\n"
                   " The computer cannot always be trusted to perform bin mergers\n"
                   " automatically, so during merging you may be shown a 3D plot\n"
                   " which should help YOU determine whether or not the bins should\n"
                   " be merged. Look carefully at each plot and then close the plot\n"
                   " to continue with the merging operation.\n"
                   " The image on the far right shows the bins after merging\n"
                   " Press any key to produce plots...")
        print "****************************************************************"

    def printSplitInstructions(self):
        raw_input( "****************************************************************\n"
                   " SPLITTING INSTRUCTIONS - PLEASE READ CAREFULLY\n"
                   "****************************************************************\n"
                   " The computer cannot always be trusted to perform bin splits\n"
                   " automatically, so during splitting you may be shown a 3D plot\n"
                   " which should help YOU determine whether or not the bin should\n"
                   " be split. Look carefully at each plot and then close the plot\n"
                   " to continue with the splitting operation.\n\n"
                   " Press any key to produce plots...")
        print "****************************************************************"

    def getPlotterMergeIds(self):
        """Prompt the user for ids to be merged and check that it's all good"""
        input_not_ok = True
        ret_bids = []
        while(input_not_ok):
            ret_bids = []
            option = raw_input("Please enter 'space' separated bin Ids or 'q' to quit: ")
            if(option.upper() == 'Q'):
                return []
            bids = option.split(" ")
            for bid in bids:
                try:
                    # check that it's an int
                    i_bid = int(bid)
                    # check that it's in the bins list
                    if(i_bid not in self.bins):
                        print "**Error: bin",bid,"not found"
                        input_not_ok = True
                        break
                    input_not_ok = False
                    ret_bids.append(i_bid)
                except ValueError:
                    print "**Error: invalid value:", bid
                    input_not_ok = True
                    break
        return ret_bids

    def promptOnMerge(self, bids=[], minimal=False):
        """Check that the user is ok with this merge"""
        input_not_ok = True
        valid_responses = ['Y','N','Q']
        vrs = ",".join([str.lower(str(x)) for x in valid_responses])
        bin_str = ""
        if(len(bids) != 0):
            bin_str = ": "+str(bids[0])
            for i in range(1, len(bids)):
                bin_str += " and "+str(bids[i])
        while(input_not_ok):
            if(minimal):
                option = raw_input(" Merge? ("+vrs+") : ")
            else:
                option = raw_input(" ****WARNING**** About to merge bins"+bin_str+"\n" \
                                   " If you continue you *WILL* overwrite existing bins!\n" \
                                   " You have been shown a 3d plot of the bins to be merged.\n" \
                                   " Continue only if you're sure this is what you want to do!\n" \
                                   " y = yes, n = no, q = no and quit merging\n" \
                                   " Merge? ("+vrs+") : ")
            if(option.upper() in valid_responses):
                print "****************************************************************"
                return option.upper()
            else:
                print "Error, unrecognised choice '"+option.upper()+"'"
                minimal = True

    def promptOnSplit(self, parts, mode, minimal=False):
        """Check that the user is ok with this split"""
        input_not_ok = True
        valid_responses = ['Y','N','C','K','P']
        vrs = ",".join([str.lower(str(x)) for x in valid_responses])
        while(input_not_ok):
            if(minimal):
                option = raw_input(" Split? ("+vrs+") : ")
            else:
                option = raw_input(" ****WARNING**** About to split bin into "+str(parts)+" parts\n" \
                                   " If you continue you *WILL* overwrite existing bins!\n" \
                                   " You have been shown a 3d plot of the bin after splitting.\n" \
                                   " Continue only if you're sure this is what you want to do!\n" \
                                   " y = yes, n = no, c = redo but use coverage profile,\n" \
                                   " k = redo but use kmer profile, p = choose new number of parts\n" \
                                   " Split? ("+vrs+") : ")
            if(option.upper() in valid_responses):
                if(option.upper() == 'K' and mode.upper() == 'KMER' or option.upper() == 'C' and mode.upper() == 'COV'):
                    print "Error, you are already using that profile to split!"
                    minimal=True
                else:
                    print "****************************************************************"
                    return option.upper()
            else:
                print "Error, unrecognised choice '"+option.upper()+"'"
                minimal = True

    def promptOnDelete(self, bids, minimal=False):
        """Check that the user is ok with this split"""
        input_not_ok = True
        valid_responses = ['Y','N']
        vrs = ",".join([str.lower(str(x)) for x in valid_responses])
        bids_str = ",".join([str.lower(str(x)) for x in bids])
        while(input_not_ok):
            if(minimal):
                option = raw_input(" Delete? ("+vrs+") : ")
            else:
                option = raw_input(" ****WARNING**** About to delete bin(s):\n" \
                                   " "+bids_str+"\n" \
                                   " If you continue you *WILL* overwrite existing bins!\n" \
                                   " Continue only if you're sure this is what you want to do!\n" \
                                   " y = yes, n = no\n"\
                                   " Delete? ("+vrs+") : ")
            if(option.upper() in valid_responses):
                print "****************************************************************"
                return option.upper()
            else:
                print "Error, unrecognised choice '"+option.upper()+"'"
                minimal = True

#------------------------------------------------------------------------------
# BIN STATS
    def findCoreCentres(self, gc_range=None, getKVals=False, processChimeric=False):
        """Find the point representing the centre of each core"""
        bin_centroid_points = np_array([])
        bin_centroid_colors = np_array([])
        bin_centroid_kvals = np_array([])
        bin_centroid_gc = np_array([])
        bids = np_array([])

        if gc_range is not None:
            # we only want to plot a subset of these guys
            gc_low = gc_range[0]
            gc_high = gc_range[1]
        num_added = 0
        for bid in self.getBids():
            if not processChimeric and self.PM.isLikelyChimeric[bid]:
                continue

            add_bin = True
            if gc_range is not None:
                avg_gc = np_median([self.PM.contigGCs[row_index] for row_index in self.bins[bid].rowIndices])
                if avg_gc < gc_low or avg_gc > gc_high:
                    add_bin = False
            if add_bin:
                bin_centroid_points = np_append(bin_centroid_points,
                                                self.bins[bid].covMedians)

                bin_centroid_colors = np_append(bin_centroid_colors, self.PM.colorMapGC(np_median(self.PM.contigGCs[self.bins[bid].rowIndices])))

                bin_centroid_gc = np_append(bin_centroid_gc, np_median(self.PM.contigGCs[self.bins[bid].rowIndices]))

                if getKVals:
                    bin_centroid_kvals = np_append(bin_centroid_kvals,
                                                   np_median([
                                                            self.PM.kmerNormPC1[row_index] for row_index in
                                                            self.bins[bid].rowIndices
                                                            ],
                                                           axis=0)
                                                   )

                bids = np_append(bids, bid)
                num_added += 1

        if num_added != 0:
            bin_centroid_points = np_reshape(bin_centroid_points, (num_added, 3))
            bin_centroid_colors = np_reshape(bin_centroid_colors, (num_added, 4))

        if getKVals:
            return (bin_centroid_points, bin_centroid_colors, bin_centroid_gc, bin_centroid_kvals, bids)

        return (bin_centroid_points, bin_centroid_colors, bin_centroid_gc, bids)

    def getAngleBetween(self, rowIndex1, rowIndex2, ):
        """Find the angle between two contig's coverage vectors"""
        u = self.PM.covProfiles[rowIndex1]
        v = self.PM.covProfiles[rowIndex2]
        try:
            ac = np_arccos(np_dot(u,v)/self.PM.normCoverages[rowIndex1]/self.PM.normCoverages[rowIndex1])
        except FloatingPointError:
            return 0
        return ac

    def scoreContig(self, rowIndex, bid):
        """Determine how well a particular contig fits with a bin"""
        return self.getBin(bid).scoreProfile(self.PM.kmerNormPC1[rowIndex], self.PM.transformedCP[rowIndex])

    def measureBinVariance(self, mode='kmer', makeKillList=False, tolerance=1.0, verbose=False):
        """Get the stats on M's across all bins

        If specified, will return a list of all bins which
        fall outside of the average M profile
        """
        Ms = {}
        Ss = {}
        Rs = {}
        for bid in self.getBids():
            if(mode == 'kmer'):
                (Ms[bid], Ss[bid], Rs[bid]) = self.bins[bid].getInnerVariance(self.PM.kmerNormPC1)
            elif(mode == 'cov'):
                (Ms[bid], Ss[bid], Rs[bid]) = self.bins[bid].getInnerVariance(self.PM.transformedCP, mode="cov")

        # find the mean and stdev
        if(not makeKillList):
            return (np_mean(np_array(Ms.values())), np_std(np_array(Ms.values())), np_median(np_array(Ss.values())), np_std(np_array(Ss.values())))

        else:
            cutoff = np_mean(np_array(Ms.values())) + tolerance * np_std(np_array(Ms.values()))
            kill_list = []
            for bid in Ms:
                if(Ms[bid] > cutoff):
                    kill_list.append(bid)
            return (kill_list, cutoff)

    def analyseBinKVariance(self, outlierTrim=0.1, plot=False):
        """Measure within and between bin variance of kmer sigs

        return a list of potentially confounding kmer indices
        """
        print "    Measuring kmer type variances"
        means = np_array([])
        stdevs = np_array([])
        bids = np_array([])

        # work out the mean and stdev for the kmer sigs for each bin
        for bid in self.getBids():
            bkworking = np_array([])
            for row_index in self.bins[bid].rowIndices:
                bkworking = np_append(bkworking, self.PM.kmerSigs[row_index])
            bkworking = np_reshape(bkworking, (self.bins[bid].binSize, np_size(self.PM.kmerSigs[0])))
            bids = np_append(bids, [bid])
            means = np_append(means, np_mean(bkworking, axis=0))
            stdevs = np_append(stdevs, np_std(bkworking, axis=0))

        means = np_reshape(means, (len(self.bins), np_size(self.PM.kmerSigs[0])))
        stdevs = np_reshape(stdevs, (len(self.bins), np_size(self.PM.kmerSigs[0])))

        # now work out the between and within core variances
        between = np_std(means, axis=0)
        within = np_median(stdevs, axis=0)

        B = np_arange(0, np_size(self.PM.kmerSigs[0]), 1)
        names = self.PM.getMerColNames().split(',')

        # we'd like to find the indices of the worst 10% for each type so we can ignore them
        # specifically, we'd like to remove the least variable between core kms and the
        # most variable within core kms.
        sort_between_indices = np_argsort(between)
        sort_within_indices = np_argsort(within)[::-1]
        number_to_trim = int(outlierTrim* float(np_size(self.PM.kmerSigs[0])))

        return_indices =[]
        for i in range(0,number_to_trim):
            if(sort_between_indices[i] not in return_indices):
                return_indices.append(sort_between_indices[i])
            if(sort_within_indices[i] not in return_indices):
                return_indices.append(sort_within_indices[i])

        if(plot):
            print "BETWEEN"
            for i in range(0,number_to_trim):
                print names[sort_between_indices[i]]
            print "WITHIN"
            for i in range(0,number_to_trim):
                print names[sort_within_indices[i]]

            plt.figure(1)
            plt.subplot(211)
            plt.plot(B, between, 'r--', B, within, 'b--')
            plt.xticks(B, names, rotation=90)
            plt.grid()
            plt.subplot(212)
            ratio = between/within
            plt.plot(B, ratio, 'r--')
            plt.xticks(B, names, rotation=90)
            plt.grid()
            plt.show()

        return return_indices

#------------------------------------------------------------------------------
# IO and IMAGE RENDERING

    def printBins(self, outFormat, fileName=""):
        """Wrapper for print handles piping to file or stdout"""
        if("" != fileName):
            try:
                # redirect stdout to a file
                stdout = open(fileName, 'w')
                self.printInner(outFormat, stdout)
            except:
                print "Error diverting stout to file:", fileName, exc_info()[0]
                raise
        else:
            self.printInner(outFormat)

    def printInner(self, outFormat, stream=sys_stdout):
        """Print bin information to STDOUT"""
        # handle the headers first
        separator = "\t"
        if(outFormat == 'contigs'):
            stream.write(separator.join(["#\"bid\"","\"cid\"","\"length\"","\"GC\""])+"\n")
        elif(outFormat == 'bins'):
            header = ["\"bin id\"","\"Likely chimeric\"","\"length (bp)\"","\"# seqs\"","\"GC mean\"","\"GC std\""]
            for i in xrange(0, len(self.PM.covProfiles[0])):
                header.append("\"Coverage " + str(i+1) + " mean\"")
                header.append("\"Coverage " + str(i+1) + " std\"")
            stream.write(separator.join(header) + "\n")
        elif(outFormat == 'full'):
            pass
        else:
            print "Error: Unrecognised format:", outFormat
            return

        for bid in self.getBids():
            self.bins[bid].makeBinDist(self.PM.transformedCP, self.PM.averageCoverages, self.PM.kmerNormPC1,
                                        self.PM.kmerPCs, self.PM.contigGCs, self.PM.contigLengths)
            self.bins[bid].printBin(self.PM.contigNames, self.PM.covProfiles, self.PM.contigGCs,
                                    self.PM.contigLengths, self.PM.isLikelyChimeric,
                                    outFormat=outFormat, separator=separator, stream=stream)

    def plotProfileDistributions(self):
        """Plot the coverage and kmer distributions for each bin"""
        for bid in self.getBids():
            self.bins[bid].plotProfileDistributions(self.PM.transformedCP, self.PM.kmerSigs, fileName="PROFILE_"+str(bid))

    def plotSelectBins(self,
                       bids,
                       plotMers=False,
                       fileName="",
                       plotEllipsoid=False,
                       ignoreContigLengths=False,
                       ET=None):
        """Plot a selection of bids in a window"""
        if plotEllipsoid and ET == None:
            ET = EllipsoidTool()

        # we need to do some fancy-schmancy stuff at times!
        if plotMers:
            num_cols = 2
        else:
            num_cols = 1

        fig = plt.figure(figsize=(6.5*num_cols, 6.5))
        ax = fig.add_subplot(1,num_cols,1, projection='3d')
        for i, bid in enumerate(bids):
            self.bins[bid].plotOnAx(ax,
                           self.PM.transformedCP,
                           self.PM.contigGCs,
                           self.PM.contigLengths,
                           self.PM.colorMapGC,
                           self.PM.isLikelyChimeric,
                           ET=ET,
                           printID=True,
                           ignoreContigLengths=ignoreContigLengths,
                           plotColorbar=(num_cols==1 and i==0)
                           )
        if plotMers:

            title = "Bin: %d : %d contigs : %s BP\n" % (bid, len(self.bins[bid].rowIndices), np_sum(self.PM.contigLengths[self.bins[bid].rowIndices]))
            title += "Coverage centroid: %d %d %d\n" % (np_median(self.PM.transformedCP[self.bins[bid].rowIndices,0]),
                                                                        np_median(self.PM.transformedCP[self.bins[bid].rowIndices,1]),
                                                                        np_median(self.PM.transformedCP[self.bins[bid].rowIndices,2]))
            title += "GC: median: %.4f stdev: %.4f\n" % (np_median(self.PM.contigGCs[self.bins[bid].rowIndices]), np_std(self.PM.contigGCs[self.bins[bid].rowIndices]))

            if self.PM.isLikelyChimeric[bid]:
                title += "Likely Chimeric"

            ax.set_title(title)

            ax = fig.add_subplot(1, 2, 2)
            for i, bid in enumerate(bids):
                self.bins[bid].plotMersOnAx(ax,
                                            self.PM.kmerPCs[:,0],
                                            self.PM.kmerPCs[:,1],
                                            self.PM.contigGCs,
                                            self.PM.contigLengths,
                                            self.PM.colorMapGC,
                                            ET=ET,
                                            printID=True,
                                            plotColorbar=(i==0)
                                            )
            ax.set_title('PCA of k-mer signature')

        fig.set_size_inches(6*num_cols, 6)
        if(fileName != ""):
            try:
                plt.savefig(fileName,dpi=300)
            except:
                print "Error saving image:", fileName, exc_info()[0]
                raise
        else:
            try:
                plt.show()
            except:
                print "Error showing image:", exc_info()[0]
                raise

        plt.close(fig)
        del fig

    def plotMultipleBins(self,
                         bins,
                         untransformed=False,
                         semi_untransformed=False,
                         ignoreContigLengths=False,
                         squash=False):
        """plot a bunch of bins, used mainly for debugging"""
        ET = EllipsoidTool()
        fig = plt.figure()

        if untransformed or semi_untransformed:
            coords = self.PM.covProfiles
            et = None
            pc = False
        else:
            coords = self.PM.transformedCP
            et = ET
            pc = True


        if squash:
            # mix all the points together
            # we need to work out how to shape the plots
            num_plots = len(bins)
            plot_rows = float(int(np_sqrt(num_plots)))
            plot_cols = np_ceil(float(num_plots)/plot_rows)
            plot_num = 1
            for bids in bins:
                ax = fig.add_subplot(plot_rows, plot_cols, plot_num, projection='3d')
                disp_vals = np_array([])
                disp_lens = np_array([])
                gcs = np_array([])
                num_points = 0
                for bid in bids:
                    for row_index in self.bins[bid].rowIndices:
                        num_points += 1
                        disp_vals = np_append(disp_vals, coords[row_index])
                        disp_lens = np_append(disp_lens, np_sqrt(self.PM.contigLengths[row_index]))
                        gcs = np_append(gcs, self.PM.contigGCs[row_index])

                # reshape
                disp_vals = np_reshape(disp_vals, (num_points, 3))

                if ignoreContigLengths:
                    sc = ax.scatter(disp_vals[:,0], disp_vals[:,1], disp_vals[:,2], edgecolors='none', c=gcs, cmap=self.PM.colorMapGC, vmin=0.0, vmax=1.0, s=10., marker='.')
                else:
                    sc = ax.scatter(disp_vals[:,0], disp_vals[:,1], disp_vals[:,2], edgecolors='k', c=gcs, cmap=self.PM.colorMapGC, vmin=0.0, vmax=1.0, s=disp_lens, marker='.')
                sc.set_edgecolors = sc.set_facecolors = lambda *args:None # disable depth transparency effect

                plot_num += 1
                
            cbar = plt.colorbar(sc, shrink=0.5)
            cbar.ax.tick_params()
            cbar.ax.set_title("% GC", size=10)
            cbar.set_ticks([0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])
            cbar.ax.set_ylim([0.15, 0.85])
            cbar.outline.set_ydata([0.15] * 2 + [0.85] * 4 + [0.15] * 3)
        else:
            # plot all separately
            # we need to work out how to shape the plots
            num_plots = len(bins) + 1
            plot_rows = float(int(np_sqrt(num_plots)))
            plot_cols = np_ceil(float(num_plots)/plot_rows)
            plot_num = 1
            ax = fig.add_subplot(plot_rows, plot_cols, plot_num, projection='3d')

            for bids in bins:
                for bid in bids:
                    self.bins[bid].plotOnAx(ax, coords, self.PM.contigGCs, self.PM.contigLengths, self.PM.colorMapGC, self.PM.isLikelyChimeric, ignoreContigLengths=ignoreContigLengths, ET=et, plotCentroid=pc)

            plot_num += 1
            if semi_untransformed:
                coords = self.PM.transformedCP
                et = ET
                pc = True

            for bids in bins:
                ax = fig.add_subplot(plot_rows, plot_cols, plot_num, projection='3d')
                plot_num += 1
                for bid in bids:
                    self.bins[bid].plotOnAx(ax, coords, self.PM.contigGCs, self.PM.contigLengths, self.PM.colorMapGC, self.PM.isLikelyChimeric, ignoreContigLengths=ignoreContigLengths, ET=et, plotCentroid=pc)

        try:
            plt.show()
        except:
            print "Error showing image:", exc_info()[0]
            raise

        plt.close(fig)
        del fig


    def plotBins(self,
                 FNPrefix="BIN",
                 sideBySide=False,
                 folder='',
                 plotEllipsoid=False,
                 ignoreContigLengths=False,
                 ET=None):
        """Make plots of all the bins"""
        if plotEllipsoid and ET == None:
            ET = EllipsoidTool()

        if folder != '':
            makeSurePathExists(folder)

        for bid in self.getBids():
            self.bins[bid].makeBinDist(self.PM.transformedCP, self.PM.averageCoverages, self.PM.kmerNormPC1, self.PM.kmerPCs, self.PM.contigGCs, self.PM.contigLengths)

        if(sideBySide):
            print "Plotting side by side"
            self.plotSideBySide(self.bins.keys(), tag=FNPrefix, ignoreContigLengths=ignoreContigLengths)
        else:
            print "Plotting bins"
            for bid in self.getBids():
                if folder != '':
                    self.bins[bid].plotBin(self.PM.transformedCP, self.PM.contigGCs, self.PM.kmerNormPC1,
                                            self.PM.contigLengths, self.PM.colorMapGC, self.PM.isLikelyChimeric,
                                            fileName=osp_join(folder, FNPrefix+"_"+str(bid)),
                                            ignoreContigLengths=ignoreContigLengths, ET=ET)
                else:
                    self.bins[bid].plotBin(self.PM.transformedCP, self.PM.contigGCs, self.PM.kmerNormPC1,
                                              self.PM.contigLengths, self.PM.colorMapGC, self.PM.isLikelyChimeric,
                                              fileName=FNPrefix+"_"+str(bid), ignoreContigLengths=ignoreContigLengths, ET=ET)
                    
    def plotBinCoverage(self, plotEllipses=False, plotContigLengs=False, printID=False):
        """Make plots of all the bins"""

        print "Plotting first 3 stoits in untransformed coverage space"
         
        # plot contigs in coverage space
        fig = plt.figure()
  
        if plotContigLengs:
            disp_lens = np_sqrt(self.PM.contigLengths)
        else:
            disp_lens = 30
            
        # plot contigs in kmer space
        ax = fig.add_subplot(121, projection='3d')
        ax.set_xlabel('kmer PC1')
        ax.set_ylabel('kmer PC2')
        ax.set_zlabel('kmer PC3')
        ax.set_title('kmer space')
        
        sc = ax.scatter(self.PM.kmerPCs[:,0], self.PM.kmerPCs[:,1], self.PM.kmerPCs[:,2], edgecolors='k', c=self.PM.contigGCs, cmap=self.PM.colorMapGC, vmin=0.0, vmax=1.0, s=disp_lens)
        sc.set_edgecolors = sc.set_facecolors = lambda *args:None # disable depth transparency effect
        
        if plotEllipses:
            ET = EllipsoidTool()
            for bid in self.getBids():
                row_indices = self.bins[bid].rowIndices
                (center, radii, rotation) = self.bins[bid].getBoundingEllipsoid(self.PM.kmerPCs[:, 0:3], ET=ET)
                centroid_gc = np_mean(self.PM.contigGCs[row_indices])
                centroid_color = self.PM.colorMapGC(centroid_gc)
                if printID:
                    ET.plotEllipsoid(center, radii, rotation, ax=ax, plotAxes=False, cageColor=centroid_color, label=self.id)
                else:
                    ET.plotEllipsoid(center, radii, rotation, ax=ax, plotAxes=False, cageColor=centroid_color)
        
        # plot contigs in untransformed coverage space
        ax = fig.add_subplot(122, projection='3d')
        ax.set_xlabel('coverage 1')
        ax.set_ylabel('coverage 2')
        ax.set_zlabel('coverage 3')
        ax.set_title('coverage space')
        
        sc = ax.scatter(self.PM.covProfiles[:,0], self.PM.covProfiles[:,1], self.PM.covProfiles[:,2], edgecolors='k', c=self.PM.contigGCs, cmap=self.PM.colorMapGC, vmin=0.0, vmax=1.0, s=disp_lens)
        sc.set_edgecolors = sc.set_facecolors = lambda *args:None # disable depth transparency effect
        
        cbar = plt.colorbar(sc, shrink=0.5)
        cbar.ax.tick_params()
        cbar.ax.set_title("% GC", size=10)
        cbar.set_ticks([0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])
        cbar.ax.set_ylim([0.15, 0.85])
        cbar.outline.set_ydata([0.15] * 2 + [0.85] * 4 + [0.15] * 3)

        if plotEllipses:
            ET = EllipsoidTool()
            for bid in self.getBids():
                row_indices = self.bins[bid].rowIndices
                (center, radii, rotation) = self.bins[bid].getBoundingEllipsoid(self.PM.covProfiles[:, 0:3], ET=ET)
                centroid_gc = np_mean(self.PM.contigGCs[row_indices])
                centroid_color = self.PM.colorMapGC(centroid_gc)
                if printID:
                    ET.plotEllipsoid(center, radii, rotation, ax=ax, plotAxes=False, cageColor=centroid_color, label=self.id)
                else:
                    ET.plotEllipsoid(center, radii, rotation, ax=ax, plotAxes=False, cageColor=centroid_color)

        try:
            plt.show()
            plt.close(fig)
        except:
            print "Error showing image", exc_info()[0]
            raise
        
        del fig
            
    def plotSideBySide(self, bids, fileName="", tag="", use_elipses=True, ignoreContigLengths=False):
        """Plot two bins side by side in 3d"""
        if use_elipses:
            ET = EllipsoidTool()
        else:
            ET = None
        fig = plt.figure()

        # get plot extents
        xMin = 1e6
        xMax = 0
        yMin = 1e6
        yMax = 0
        zMin = 1e6
        zMax = 0

        for bid in bids:
            x = self.PM.transformedCP[self.bins[bid].rowIndices,0]
            y = self.PM.transformedCP[self.bins[bid].rowIndices,1]
            z = self.PM.transformedCP[self.bins[bid].rowIndices,2]

            xMin = min(min(x), xMin)
            xMax = max(max(x), xMax)

            yMin = min(min(y), yMin)
            yMax = max(max(y), yMax)

            zMin = min(min(z), zMin)
            zMax = max(max(z), zMax)

        # we need to work out how to shape the plots
        num_plots = len(bids)
        plot_rows = float(int(np_sqrt(num_plots)))
        plot_cols = np_ceil(float(num_plots)/plot_rows)
        for plot_num, bid in enumerate(bids):
            title = self.bins[bid].plotOnFig(fig, plot_rows, plot_cols, plot_num+1,
                                              self.PM.transformedCP, self.PM.contigGCs, self.PM.contigLengths,
                                              self.PM.colorMapGC, self.PM.isLikelyChimeric, ET=ET, fileName=fileName,
                                              plotColorbar=(plot_num == len(bids)-1), extents=[xMin, xMax, yMin, yMax, zMin, zMax],
                                              ignoreContigLengths=ignoreContigLengths)

            plt.title(title)
        if(fileName != ""):
            try:
                fig.set_size_inches(12,6)
                plt.savefig(fileName,dpi=300)
            except:
                print "Error saving image:", fileName, exc_info()[0]
                raise
        elif(show):
            try:
                plt.show()
            except:
                print "Error showing image:", exc_info()[0]
                raise
        plt.close(fig)
        del fig

    def plotStoitNames(self, ax):
        """Plot stoit names on an existing axes"""
        self.PM.plotStoitNames(ax)

    def plotBinIds(self, gc_range=None, ignoreRanges=False, showChimeric=False):
        """Render 3d image of core ids"""
        (bin_centroid_points, bin_centroid_colors, _bin_centroid_gc, bids) = self.findCoreCentres(gc_range=gc_range,
                                                                                                  processChimeric=showChimeric)
        fig = plt.figure()
        ax = fig.gca(projection='3d')
        ax.set_xlabel('x coverage')
        ax.set_ylabel('y coverage')
        ax.set_zlabel('z coverage')

        outer_index = 0
        for bid in bids:
            ax.text(bin_centroid_points[outer_index,0],
                    bin_centroid_points[outer_index,1],
                    bin_centroid_points[outer_index,2],
                    str(int(bid)),
                    color=bin_centroid_colors[outer_index]
                    )
            outer_index += 1

        if ignoreRanges:
            mm = np_max(bin_centroid_points, axis=0)
            ax.set_xlim3d(0, mm[0])
            ax.set_ylim3d(0, mm[1])
            ax.set_zlim3d(0, mm[2])

        else:
            self.plotStoitNames(ax)
            ax.set_xlim3d(0, 1000)
            ax.set_ylim3d(0, 1000)
            ax.set_zlim3d(0, 1000)
        try:
            plt.show()
            plt.close(fig)
        except:
            print "Error showing image", exc_info()[0]
            raise
        del fig

    def plotBinPoints(self, ignoreRanges=False, plotColorbar=True, showChimeric=False):
        """Render the image for validating cores"""
        (bin_centroid_points, _bin_centroid_colors, bin_centroid_gc, _bids) = self.findCoreCentres(processChimeric=showChimeric)
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        print bin_centroid_gc
        sc = ax.scatter(bin_centroid_points[:,0], bin_centroid_points[:,1], bin_centroid_points[:,2], edgecolors='k', c=bin_centroid_gc, cmap=self.PM.colorMapGC, vmin=0.0, vmax=1.0)
        sc.set_edgecolors = sc.set_facecolors = lambda *args:None # disable depth transparency effect

        if plotColorbar:
            cbar = plt.colorbar(sc, shrink=0.5)
            cbar.ax.tick_params()
            cbar.ax.set_title("% GC", size=10)
            cbar.set_ticks([0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])
            cbar.ax.set_ylim([0.15, 0.85])
            cbar.outline.set_ydata([0.15] * 2 + [0.85] * 4 + [0.15] * 3)

        ax.set_xlabel('x coverage')
        ax.set_ylabel('y coverage')
        ax.set_zlabel('z coverage')

        if not ignoreRanges:
            self.plotStoitNames(ax)
            ax.set_xlim3d(0, 1000)
            ax.set_ylim3d(0, 1000)
            ax.set_zlim3d(0, 1000)
        try:
            plt.show()
            plt.close(fig)
        except:
            print "Error showing image", exc_info()[0]
            raise
        del fig

    def plotTDist(self, scores, testPoint):
        """DEBUG: Plot the distribution of the NULL T disytribution"""
        B = np_arange(0, len(scores), 1)
        co_90 = [int(float(len(scores))*0.90)] * 100
        co_95 = [int(float(len(scores))*0.95)] * 100
        co_97 = [int(float(len(scores))*0.97)] * 100
        co_99 = [int(float(len(scores))*0.99)] * 100
        smin = np_min(scores)
        smax = np_max(scores)
        step = (smax-smin)/100
        LL = np_arange(smin, smax, step)

        fig = plt.figure()

        plt.plot(co_90, LL, 'g-')
        plt.plot(co_95, LL, 'g-')
        plt.plot(co_97, LL, 'g-')
        plt.plot(co_99, LL, 'g-')

        plt.plot(B, scores, 'r-')
        plt.plot(B, [testPoint]*len(scores), 'b-')
        plt.show()
        plt.close(fig)
        del fig

###############################################################################
###############################################################################
###############################################################################
###############################################################################



#!/usr/bin/env python
###############################################################################
#                                                                             #
#    rainbow.py                                                               #
#                                                                             #
#    Simple dimple heatmap                                                    #
#                                                                             #
#    Copyright (C) Michael Imelfort                                           #
#                                                                             #
###############################################################################
#                                                                             #
#          .d8888b.                                    888b     d888          #
#         d88P  Y88b                                   8888b   d8888          #
#         888    888                                   88888b.d88888          #
#         888        888d888 .d88b.   .d88b.  88888b.  888Y88888P888          #
#         888  88888 888P"  d88""88b d88""88b 888 "88b 888 Y888P 888          #
#         888    888 888    888  888 888  888 888  888 888  Y8P  888          #
#         Y88b  d88P 888    Y88..88P Y88..88P 888 d88P 888   "   888          #
#          "Y8888P88 888     "Y88P"   "Y88P"  88888P"  888       888          #
#                                             888                             #
#                                             888                             #
#                                             888                             #
#                                                                             #
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Michael Imelfort"
__copyright__ = "Copyright 2012/2013"
__credits__ = ["Michael Imelfort"]
__license__ = "GPL3"
__version__ = "0.2.1"
__maintainer__ = "Michael Imelfort"
__email__ = "mike@mikeimelfort.com"
__status__ = "Alpha"

###############################################################################


import math


###############################################################################
###############################################################################
###############################################################################
###############################################################################

class Rainbow:
    def __init__(self, lb, ub, res, type="rb"):
        """
        Specify the upper and lower bounds for your data. 
        resolution refers to the number of bins which are available in this space
        Supports four heatmap types is red-blue, blue-red, red-green-blue and blue-green-red
        """ 

        # constants
        self.RB_lower_offset = 0.5
        self.RB_divisor = (2.0/3.0)
        self.RB_ERROR_COLOUR = [0,0,0]

        # set the limits
        self.lowerBound = lb
        self.upperBound = ub
        self.resolution = res
        self.tickSize = (self.upperBound - self.lowerBound)/(self.resolution - 1)
        
        # set the type, red-blue by default
        self.type = type
        self.redOffset = 0.0
        self.greenOffset = self.RB_divisor * math.pi * 2.0
        self.blueOffset = self.RB_divisor * math.pi
        
        self.ignoreRed = False
        self.ignoreGreen = True
        self.ignoreBlue = False
        
        self.lowerScale = 0.0
        self.upperScale = self.RB_divisor * math.pi
        
        if(self.type == "rbg"): # red-blue-green
            self.redOffset = 0.0
            self.greenOffset = self.RB_divisor * math.pi * 2.0
            self.blueOffset = self.RB_divisor * math.pi
            
            self.ignoreRed = False
            self.ignoreGreen = False
            self.ignoreBlue = False
            
            self.lowerScale = 0.0
            self.upperScale = (self.RB_divisor * math.pi * 2.0)

        elif(self.type == "gbr"): # green-blue-red
            self.redOffset = self.RB_divisor * math.pi * 2.0
            self.greenOffset = 0.0
            self.blueOffset = self.RB_divisor * math.pi

            self.ignoreRed = False
            self.ignoreGreen = False
            self.ignoreBlue = False
            
            self.lowerScale = 0.0
            self.upperScale = (self.RB_divisor * math.pi * 2.0)

        elif(self.type == "br"): # blue-red
            self.redOffset = self.RB_divisor * math.pi
            self.greenOffset = self.RB_divisor * math.pi * 2.0
            self.blueOffset = 0.0
            
            self.ignoreRed = False
            self.ignoreGreen = True
            self.ignoreBlue = False

            self.lowerScale = 0.0
            self.upperScale = (self.RB_divisor * math.pi)

        self.scaleMultiplier = (self.upperScale - self.lowerScale)/(self.upperBound - self.lowerBound)
    
    def getValue(self, val):
        """Get a raw value, not a color"""
        return (math.cos(val) + self.RB_lower_offset) * self.RB_divisor

    def getColor(self, val):
        """Return a color for the given value.
        
        If nothing makes sense. return black
        """
        if(val > self.upperBound or val < self.lowerBound):
            return self.RB_ERROR_COLOUR
        
        # normalise the value to suit the ticks
        normalised_value = round(val/self.tickSize) * self.tickSize
    
        # map the normalised value onto the horizontal scale
        scaled_value = (normalised_value - self.lowerBound) * self.scaleMultiplier + self.lowerScale
            
        red = 0
        green = 0
        blue = 0
        
        if(not self.ignoreRed):
            red = int(round(self.getValue(scaled_value - self.redOffset) * 255))
        if(not self.ignoreGreen):
            green = int(round(self.getValue(scaled_value - self.greenOffset) * 255))
        if(not self.ignoreBlue):
            blue = int(round(self.getValue(scaled_value - self.blueOffset) * 255))
    
        return (red, green, blue)

###############################################################################
###############################################################################
###############################################################################
###############################################################################
